#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that contains windows only functions

@package burger._find_visual_studio

@var burger._find_visual_studio._VSTOOLS
List of Visual Studio tools

@var burger._find_visual_studio._IDETOOLS
List of IDE tools

@var burger._find_visual_studio._VSVERSIONS
Versions of Visual Studio from 2003 to 2015

@var burger._find_visual_studio._SUPPORTED_CPUS
List of supported CPUs for Visual Studio  tools
"""

# pylint: disable=useless-object-inheritance
# pylint: disable=too-many-arguments
# pylint: disable=unnecessary-pass
# pylint: disable=too-few-public-methods

from __future__ import absolute_import, print_function, unicode_literals

import os.path

from ._vsinstance import VisualStudioInstance

try:
    from wslwinreg import convert_from_windows_path, get_file_info
except ImportError:
    pass


# List of Visual Studio tools
_VSTOOLS = ("cl.exe", "link.exe", "lib.exe")

# List of IDE tools
_IDETOOLS = ("devenv.exe", "devenv.com")

# Versions of Visual Studio from 2003 to 2015
_VSVERSIONS = ("7.1", "8.0", "9.0", "10.0", "11.0", "12.0", "14.0")

# List of supported CPUs for Visual Studio  tools
_SUPPORTED_CPUS = ("x86", "x64", "arm", "arm64")

########################################


def _find_vs2003_2015(installed_roots):
    """
    Find all versions of Visual Studio from 2003 to 2015.

    Args:
        installed_roots: Open registry key "Software\\Microsoft"
    """

    # pylint: disable=too-many-branches
    result_list = []

    for key in _VSVERSIONS:

        known_paths = {}

        # Get the root pathname to the Visual Studio version
        try:
            sub_key = installed_roots.open_subkey("VisualStudio\\SxS\\VS7")
            vs_path = convert_from_windows_path(
                sub_key.get_value(key)[0])
        except OSError:
            continue

        # Locate devenv.exe and related files
        for item in _IDETOOLS:
            test_dir = os.path.join(
                vs_path, "Common7", "IDE", item)
            if os.path.exists(test_dir):
                known_paths[item + "_x86"] = test_dir

        # vcvarsall.bat is a special case
        if key == "7.1":
            test_dir = os.path.join(
                vs_path, "Common7", "Tools", "vsvars32.bat")
            if os.path.exists(test_dir):
                known_paths["vcvarsall.bat"] = test_dir

        # Get the pathname to the Visual Studio tools
        try:
            sub_key = installed_roots.open_subkey("VisualStudio\\SxS\\VC7")
            vc7_path = convert_from_windows_path(
                sub_key.get_value(key)[0])
        except OSError:
            continue

        # vcvarsall.bat is a special case
        if key != "7.1":
            test_dir = os.path.join(vc7_path, "vcvarsall.bat")
            if os.path.exists(test_dir):
                known_paths["vcvarsall.bat"] = test_dir

        # Locate the compiler and linkers
        for item in _VSTOOLS:
            test_dir = os.path.join(vc7_path, "bin", item)
            if os.path.exists(test_dir):
                known_paths[item + "_x86"] = test_dir

        # Locate msbuild, if available. VS 2013 and higher.
        try:
            sub_key = installed_roots.open_subkey(
                "MSBuild\\ToolsVersions\\" + key)
            vc7_path = convert_from_windows_path(
                sub_key.get_value("MSBuildToolsPath")[0])
        except OSError:
            vc7_path = None

        # Is there an MSBuild toolchain?
        if vc7_path:
            test_dir = os.path.join(vc7_path, "msbuild.exe")
            if os.path.exists(test_dir):
                known_paths["msbuild.exe" + "_x86"] = test_dir
            test_dir = os.path.join(vc7_path, "amd64", "msbuild.exe")
            if os.path.exists(test_dir):
                known_paths["msbuild.exe" + "_x64"] = test_dir

        version_number = get_file_info(
            known_paths["devenv.exe_x86"], "ProductVersion")
        file_description = get_file_info(
            known_paths["devenv.exe_x86"], "FileDescription")
        result_list.append(VisualStudioInstance(
            file_description,
            version_number,
            vs_path,
            known_paths
        ))

    return result_list

########################################


def _find_vs2017_higher(installed_roots):
    """
    Find all versions of Visual Studio 2017 and higher.

    Args:
        installed_roots: Open registry key "Software\\Microsoft"
    """

    # pylint: disable=too-many-branches
    result_list = []

    # Scan registery keys for VisualStudio_*
    for key in installed_roots.get_subkeys():

        if not key.startswith("VisualStudio_"):
            continue

        # Found an installed copy of Visual Studio
        known_paths = {}

        # Get the root pathname to the Visual Studio version
        try:
            sub_key = installed_roots.open_subkey(key + "\\Capabilities")
            vs_path = sub_key.get_value("ApplicationDescription")[0]
        except OSError:
            continue

        # Get rid of extra characters
        # Example: @C:\Program Files (x86)\Microsoft Visual Studio\
        #          2017\Community\Common7\IDE\devenvdesc.dll,-1004
        vs_path = convert_from_windows_path(
            vs_path.lstrip("@").rsplit(",", 1)[0])
        vs_path = os.path.dirname(
            os.path.dirname(
                os.path.dirname(vs_path)))

        # Locate devenv.exe and related files
        for item in _IDETOOLS:
            test_dir = os.path.join(
                vs_path, "Common7", "IDE", item)
            if os.path.exists(test_dir):
                known_paths[item + "_x86"] = test_dir

        # vcvarsall.bat is a special case
        test_dir = os.path.join(
            vs_path,
            "VC",
            "Auxiliary",
            "Build",
            "vcvarsall.bat")
        if os.path.exists(test_dir):
            known_paths["vcvarsall.bat"] = test_dir

        version_number = get_file_info(
            known_paths["devenv.exe_x86"], "ProductVersion")
        file_description = get_file_info(
            known_paths["devenv.exe_x86"], "FileDescription")

        # Locate msbuild
        test_dir = os.path.join(
            vs_path,
            "MSBuild",
            version_number.split(".")[0] + ".0",
            "Bin",
            "MSBuild.exe")
        if os.path.exists(test_dir):
            known_paths["msbuild.exe" + "_x86"] = test_dir
            test_dir = os.path.dirname(test_dir)
            test_dir = os.path.join(test_dir, "amd64", "msbuild.exe")
            if os.path.exists(test_dir):
                known_paths["msbuild.exe" + "_x64"] = test_dir
        else:
            test_dir = os.path.join(
                vs_path,
                "MSBuild",
                "Current",
                "Bin",
                "MSBuild.exe")
            if os.path.exists(test_dir):
                known_paths["msbuild.exe" + "_x86"] = test_dir

            test_dir = os.path.dirname(test_dir)
            test_dir = os.path.join(test_dir, "amd64", "msbuild.exe")
            if os.path.exists(test_dir):
                known_paths["msbuild.exe" + "_x64"] = test_dir

        result_list.append(VisualStudioInstance(
            file_description,
            version_number,
            vs_path,
            known_paths
        ))

    return result_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that contains file manipulation functions

@package burger.fileutils
"""

# pylint: disable=consider-using-f-string
# pylint: disable=redefined-builtin
# pyright: reportMissingImports=false

from __future__ import absolute_import, print_function, unicode_literals

import errno
import os
import shutil
import stat
import codecs

from .strutils import is_string, convert_to_array, encapsulate_path, \
    get_windows_host_type, translate_to_regex_match, \
    IS_MACOSX, IS_WINDOWS_HOST, IS_LINUX

# Redefining built-in (Ignore redefinition of zip)
try:
    import itertools.izip as zip
except ImportError:
    pass

########################################


def is_write_protected(path_name):
    """
    Test if a file is write protected

    If the file/directory exists, it is tested if it's
    write protected. If it exists and is write protected,
    True is returned, otherwise False is returned.

    Args:
        path_name: Path name to the file/directory

    Returns:
        True if the file exists and is write protected
    """

    try:
        # Get the status of the file
        path_mode = os.stat(path_name).st_mode

        # Return True if write protected
        return not path_mode & stat.S_IWRITE

    # Don't throw if the file didn't exist
    except OSError as error:
        if error.errno != errno.ENOENT:
            raise
    return False

########################################


def make_executable(exe_path):
    """
    Set the executable flag to true on a file

    Args:
        exe_path: Pathname to the executable to fix up.

    """

    # Get the Read access bits and copy them to the X flags
    exe_mode = os.stat(exe_path).st_mode
    exe_mode |= (exe_mode & 0x124) >> 2
    os.chmod(exe_path, exe_mode)

########################################


def create_folder_if_needed(path):
    """
    Given a pathname to a folder, if the folder doesn't exist, create it.

    Call os.makedirs(path) but does not throw an
    exception if the directory already exists. All other exceptions
    are passed through with raise.

    Args:
        path: A string object with the pathname.

    See Also:
        delete_directory
    """

    try:
        os.makedirs(path)
    except OSError as error:
        if error.errno != errno.EEXIST:
            raise

########################################


def delete_file(filename):
    """
    Given a pathname to a file, delete it.

    If the file doesn't exist, it will return without raising
    an exception.

    Args:
        filename: A string object with the filename
    See Also:
        delete_directory
    """

    try:
        os.remove(filename)
    except OSError as error:
        if error.errno != errno.ENOENT:
            raise

########################################


def is_source_newer(source, destination):
    """
    Return False if the source file is older then the destination file

    Check the modification times of both files to determine if the
    source file is newer. If the destination file is older or doesn't exist
    True is returned.

    Return False if destination is newer, not False if not.

    Examples:
        result = burger.fileutils.is_source_newer("file.c", "file.obj")

        if result == 2:
            build_file_c()

        if result:
            compile("file.c", "file.obj")
        else:
            print("Already built")

    Note:
        If the source file does not exist, the function will return 2.
        This is to allow proper error checking if the source is required
        to exist.

    Args:
        source: string pathname of the file to test
        destination: string pathname of the file to test against

    Returns:
        False if not newer, True if newer, 2 if there is no source file
    """

    # Get the source file's modification time, If there's no source file,
    # return 2
    try:
        srctime = os.path.getmtime(source)
    except OSError as error:
        if error.errno != errno.ENOENT:
            raise
        return 2

    # Get the destination file's modification time, if missing return True
    try:
        desttime = os.path.getmtime(destination)
    except OSError as error:
        if error.errno != errno.ENOENT:
            raise
        return True

    # Is the source older or equal? Return False to not update the destination
    if srctime <= desttime:
        return False
    return True

########################################


def copy_file_if_needed(source, destination, verbose=True, perforce=False):
    """
    Copy a file only if newer than the destination.

    Copy a file only if the destination is missing or is older than the source
    file.

    Args:
        source: string pathname of the file to copy from
        destination: string pathname of the file to copy to
        verbose: True if print output is desired
        perforce: True if Perforce "p4 edit" should be done on the destination.

    Returns:
        Zero if no error otherwise IOError.errno

    See Also:
        is_source_newer
    """

    # If there is a destination file, check the modification times
    # and check with "is True" to ensure source exists

    if is_source_newer(source, destination) is True:

        # Alert perforce that the file is to be modified
        if perforce and is_write_protected(destination):
            # pylint: disable=import-outside-toplevel
            from .perforce import perforce_edit
            perforce_edit(destination, verbose=verbose)

        # Copy the file
        if verbose:
            print("Copying {0} -> {1}".format(source, destination))
        try:
            shutil.copyfile(source, destination)
        except IOError as error:
            print(error)
            return error.errno
    return 0

########################################


def copy_directory_if_needed(source, destination, exception_list=None,
                             verbose=True):
    """
    Copy all of the files in a directory into a new directory

    Creating any necessary directories in the process, and it
    will skip files with specific extensions

    Note:
        This is a recursive function

    Args:
        source: string pathname of the directory to copy from
        destination: string pathname of the directory to copy to
        exception_list: optional list of file extensions to ignore during copy
        verbose: True if print output is desired

    Returns:
        Zero if no error, non-zero on error

    See Also:
        copy_file_if_needed, create_folder_if_needed
    """

    # Ensure there is an exception list, even if empty
    if exception_list is None:
        exception_list = []

    # Make sure the output folder exists
    create_folder_if_needed(destination)

    # Iterate over the directory list
    for base_name in os.listdir(source):
        for item in exception_list:
            if base_name.endswith(item):
                break
        else:

            # Perform the copy of the entry
            file_name = os.path.join(source, base_name)

            # Handle the directories found
            if os.path.isdir(file_name):
                # Recursive!
                error = copy_directory_if_needed(
                    file_name, os.path.join(destination, base_name),
                    exception_list, verbose=verbose)
            else:
                error = copy_file_if_needed(file_name, os.path.join(
                    destination, base_name), verbose=verbose)

            # Exit immediately on error
            if error != 0:
                return error
    return 0

########################################


def shutil_readonly_cb(func, path, exception_info):
    """
    Subroutine for shutil.rmtree() to delete read only files

    shutil.rmtree() raises an exception if there are read
    only files in the directory being deleted. Use this
    callback to allow read only files to be disposed of.

    Examples:
        import burger
        import shutil

        shutil.rmtree(PATH_TO_DIRECTORY,onerror = burger.shutil_readonly_cb)

    Note:
        This is a callback function

    Args:
        func: Not used
        path: pathname of the file that is read only
        exception_info: Information about the exception

    See Also:
        delete_directory
    """

    # File not found? Ignore
    value = exception_info[1]
    if value.args[0] == errno.ENOENT:
        return

    # Read only?
    # EACCESS for Linux/MacOSX, EIO for Windows

    if value.args[0] == errno.EACCES or value.args[0] == errno.EIO:

        # Mark as writable and try again to delete the file
        os.chmod(path, stat.S_IWRITE)
        os.unlink(path)

########################################


def delete_directory(path, delete_read_only=False):
    """
    Recursively delete a directory

    Delete a directory and all of the files and directories within.

    Args:
        path: Pathname of the directory to delete
        delete_read_only: True if read only files are to be deleted as well

    See Also:
        shutil_readonly_cb, create_folder_if_needed
    """

    if delete_read_only:
        shutil.rmtree(path, onerror=shutil_readonly_cb)
    else:
        shutil.rmtree(path, ignore_errors=True)

########################################


def clean_directories(path, name_list, recursive=False):
    """
    Recursively clean directories with a name list

    Args:
        path: Pathname of the directory to scan
        name_list: Iterable of directory names
        recursive: Boolean if recursive clean is desired

   Examples:
        # Delete all temp and __pycache__ files recursively
        burger.fileutils.clean_directories(
            ".",
            ("*.temp", "__pycache__"),
            True)

    See Also:
        clean_files, delete_directory
    """

    match_list = translate_to_regex_match(name_list)
    for base_name in os.listdir(path):
        file_name = os.path.join(path, base_name)
        # Is it a directory? (Skip files)
        if os.path.isdir(file_name):
            for item in match_list:
                if item(base_name):
                    delete_directory(file_name)
                    break
            else:
                if recursive:
                    # Recurse if needed
                    clean_directories(file_name, name_list, recursive)

########################################


def clean_files(path, name_list, recursive=False):
    """
    Recursively clean files with a filename list

    Args:
        path: Pathname of the directory to scan
        name_list: Iterable of file names
        recursive: Boolean if recursive clean is desired

   Examples:
        # Delete all .obj and .lib files recursively
        burger.fileutils.clean_files("temp", ("*.obj", "*.lib"), True)

    See Also:
        delete_file, delete_directory
    """

    # Scan the directory
    match_list = translate_to_regex_match(name_list)
    for base_name in os.listdir(path):
        # Create full pathname
        file_name = os.path.join(path, base_name)
        # Is it a file? (Skip directories)
        if os.path.isfile(file_name):
            for item in match_list:
                if item(base_name):
                    try:
                        os.remove(file_name)
                    except OSError:
                        pass
                    break

        # Recurse if desired
        elif recursive and os.path.isdir(file_name):
            clean_files(file_name, name_list, recursive)

########################################


def get_tool_path(tool_folder, tool_name, encapsulate=False):
    """
    Find executable tool directory

    For allowing builds on multiple operating system hosts under the Burgerlib
    way of project management, it's necessary to query what is the host
    operating system and glean out which folder to find a executable compiled
    for that specific host

    Args:
        tool_folder: Pathname to the folder that contains the executables
        tool_name: Bare name of the tool (Windows will append ".exe")
        encapsulate: False if a path is requested, True if it's quoted to be
            used as a string to be sent to command line shell

    Returns:
        Full pathname to the tool to execute
    """

    # Macosx uses fat binaries
    if IS_MACOSX:
        exename = os.path.join(tool_folder, "macosx", tool_name)

    # Linux is currently just 64 bit Intel, will have to update
    # as more platforms are supported
    elif IS_LINUX:
        exename = os.path.join(tool_folder, "linux", tool_name)

    # Windows supports 32 and 64 bit Intel
    elif IS_WINDOWS_HOST:
        exename = os.path.join(
            tool_folder,
            "windows",
           	get_windows_host_type(True),
            tool_name + ".exe")

    else:

        # On unknown platforms, assume the tool is in the path for the fallback
        exename = tool_name

    # Encase in quotes to handle spaces in filenames

    if encapsulate:
        return encapsulate_path(exename)
    return exename

########################################


def traverse_directory(
        working_dir, filename_list, terminate=False, find_directory=False):
    """
    Create a list of all copies of a file following a directory

    Starting with a working directory, test if a file exists
    and if so, insert it into a list. The list will be
    starting from the root with the last entry
    being at the working directory

    Args:
        working_dir: string with the path of the folder to start the search
        filename_list: string or an iterable of strings with the name(s)
            of the file(s) to find in the scanned folders
        terminate: True if searching will end on the first found file
        find_directory: True if searching for directories instead of files.
    Returns:
        List of pathnames (With filename appended)
    """

    # Ensure that if the input was a string, that it becomes
    # an iterable to work below
    filename_list = convert_to_array(filename_list)

    # Convert into a unpacked pathname
    tempdir = os.path.abspath(working_dir)

    dirlist = []

    # Loop
    while True:
        # Iterate over the list and detect if these files are present
        for item in filename_list:
            temppath = os.path.join(tempdir, item)
            if find_directory:
                if os.path.isdir(temppath):
                    # Insert at the beginning
                    dirlist.insert(0, temppath)
                    if terminate:
                        return dirlist

            elif os.path.isfile(temppath):
                # Insert at the beginning
                dirlist.insert(0, temppath)
                if terminate:
                    return dirlist

        # Pop a folder
        tempdir2 = os.path.dirname(tempdir)
        # Already at the top of the directory?
        if tempdir2 is None or tempdir2 == tempdir:
            break
        # Use the new folder
        tempdir = tempdir2

    # Return the list of files
    return dirlist

########################################


def unlock_files(working_dir, recursive=False):
    """
    Iterate over a directory and unlock all read-only files.

    This function will generate a list of fully qualified pathnames
    of every file that was unlocked. Directories will be skipped.

    Examples:
        # Any file that is read only in this directory is now unlocked
        lock_list = unlock_files("~/projects/lockedfiles")

        # Do stuff on the files
        do_code_on_unlocked_files()

        # Re-lock all the files that were unlocked.
        lock_files(lock_list)

    Args:
        working_dir: Pathname to the directory to traverse for read-only files
        recursive: False (default) don't recurse through folders, True, recurse
    Returns:
        A list object with the name of every file that was unlocked.

    See Also:
        lock_files

    """

    # Ensure the pathname is unmangled
    abs_dir = os.path.abspath(working_dir)

    # Iterate over the directory
    result = []
    for item in os.listdir(abs_dir):
        path_name = os.path.join(abs_dir, item)

        # Get the status of the file
        path_mode = os.stat(path_name).st_mode

        # Process files
        if path_mode & stat.S_IFREG:

            # Only care about write protected files
            if not path_mode & stat.S_IWRITE:

                # Remove write protection while retaining the other flags
                flags = stat.S_IWRITE + stat.S_IWGRP + stat.S_IWOTH
                path_mode = path_mode + flags
                os.chmod(path_name, path_mode)
                result.append(path_name)
        else:
            # Process recursion
            if recursive and path_mode & stat.S_IFDIR:
                result += unlock_files(path_name, True)
    return result


########################################


def lock_files(lock_list):
    """
    Iterate over the input list and mark all files as read-only

    Args:
        lock_list: Iterable object containing a list of path names to files
            or directories to mark as "read-only"
    See Also:
        unlock_files
    """
    for item in lock_list:

        # Get the status of the file
        path_mode = os.stat(item).st_mode

        # Mark it write protected for Perforce
        mask = ~(stat.S_IWRITE + stat.S_IWGRP + stat.S_IWOTH)
        os.chmod(item, path_mode & mask)

########################################


def load_text_file(file_name):
    """
    Load in a text file as a list of lines

    Read in a text file as a list of lines and handle
    all three line ending types (\\r, \\n and \\r\\n)

    Note:
        This function assumes the file is utf-8 with or without
        a byte order mark.

    Args:
        file_name: File to load

    Returns:
        A list object with the file

    See Also:
        save_text_file, compare_files
    """

    # Open the file in a way that ensur

    try:
        with open(file_name, "rb") as filep:
            result = filep.read().decode("utf-8-sig").splitlines()
    except IOError as error:
        # Only deal with file not found
        if error.errno != errno.ENOENT:
            raise
        # If not found, return None
        result = None
    return result

########################################


def save_text_file(file_name, text_lines, line_feed=None, bom=False):
    """
    Save in a text file from an iterable of lines

    Save a text file from an iterable of lines and allow custom
    line endings. If line_feed is None, the line feed will be the
    system default.

    Note:
        This function will write out the text file using utf-8 encoding.

    Args:
        file_name: File to load
        text_lines: Lines to save
        line_feed: String to use as a line feed
        bom: If True write the UTF-8 Byte Order Mark

    See Also:
        load_text_file
    """

    # Set the proper line feed if not supplied
    if not line_feed:
        line_feed = os.linesep

    # If it's a single line, convert to an iterable
    text_lines = convert_to_array(text_lines)

    # Write out the file
    with open(file_name, "wb") as filep:

        # Write the byte order mark
        if bom:
            filep.write(codecs.BOM_UTF8)

        filep.write(line_feed.join(text_lines).encode("utf-8"))

        # Make sure there's an ending line feed
        filep.write(line_feed.encode("utf-8"))

########################################


def compare_files(filename1, filename2):
    """
    Compare text files for equality

    Check if two text files are the same length,
    and then test the contents to verify equality.

    Args:
        filename1: string object with the pathname of the file to test
        filename2: string object with the pathname of the file to test against

    Returns:
        True if the files are equal, False if not.

    See Also:
        compare_file_to_string
    """

    # Load in the two text files

    file_one_lines = load_text_file(filename1)

    # If not found, return "not equal"
    # Note: Must use is not None because empty lists are acceptable

    if file_one_lines is not None:
        file_two_lines = load_text_file(filename2)
        if file_two_lines is not None:

            # Compare the file contents
            if len(file_one_lines) == len(file_two_lines):
                for i, j in zip(file_one_lines, file_two_lines):
                    if i != j:
                        break
                else:
                    # It's a match!
                    return True
    return False

########################################


def compare_file_to_string(file_name, text_lines):
    """
    Compare text file and a string for equality

    Check if a text file is the same as a string by loading the text file and
    testing line by line to verify the equality of the contents

    Args:
        file_name: string object with the pathname of the file to test
        text_lines: string object to test against

    Returns:
        True if the file and the string are the same, False if not

    See Also:
        compare_files
    """

    # Do a data compare as a text file

    file_one_lines = load_text_file(file_name)

    # If not found, return "not equal"
    # Note: Must use is not None because empty lists are acceptable
    if file_one_lines is not None:

        # No data? Assume it's empty
        if text_lines is None:
            file_two_lines = []

        # Test if this is a StringIO object
        elif hasattr(text_lines, "getvalue"):
            file_two_lines = text_lines.getvalue().splitlines()
        # Test if a single string
        elif is_string(text_lines):
            file_two_lines = text_lines.splitlines()
        else:
            # Assume it's an iterable
            file_two_lines = text_lines

        # Compare the file contents taking into account
        # different line endings

        if len(file_one_lines) == len(file_two_lines):
            for i, j in zip(file_one_lines, file_two_lines):
                if i != j:
                    break
            else:
                # It's a match!
                return True
    return False

########################################


def read_zero_terminated_string(filep, encoding="utf-8"):
    """
    Read a zero terminated string from an open binary file

    Read in a stream of bytes and stop at the end of file or
    a terminating zero. The string will be converted from utf-8
    into unicode by default before returning.

    Args:
        filep: File record of a file opened in binary mode
        encoding: Character set encoding of the string
    Returns:
        None or the unicode string (Without the terminating zero)
    """

    # Nothing parsed yet
    chars = bytearray()
    while True:
        # Read to end of file or to a terminating character
        character = filep.read(1)

        # EOF?
        if not character:
            if chars:
                break
            return None

        # Convert to integer to handle utf-8 decoding
        temp = ord(character)

        # Zero terminator?
        if not temp:
            break
        # Add to the array
        chars.append(temp)

    # Ensure that UTF-8 data is properly parsed
    return chars.decode(encoding)

########################################


def save_text_file_if_newer(file_name, text_lines, line_feed=None,
                            bom=False, perforce=False, verbose=False):
    """
    Save in a text file from an iterable of lines if newer.

    Compare an iterable of lines to a pre-existing text file. If
    the text file either exists or differs from the input, write
    a new text file to disk.

    Note:
        This function will write out the text file using utf-8 encoding.

    Args:
        file_name: File to load
        text_lines: Lines to save
        line_feed: String to use as a line feed
        bom: If True write the UTF-8 Byte Order Mark
        perforce: Enable perforce checkout or add if True
        verbose: Enable messages if True
    Returns:
        True if no change was performed, False if the file was written

    See Also:
        save_text_file, compare_file_to_string, perforce_edit, perforce_add
    """

    # Too many arguments
    # pylint: disable=R0913

    if compare_file_to_string(file_name, text_lines):
        if verbose:
            print("{} was not changed.".format(file_name))
        return True

    # pylint: disable=import-outside-toplevel
    # Check out the file if Perforce is requested
    do_perforce_add = False
    if perforce:
        # New file?
        if not os.path.isfile(file_name):
            do_perforce_add = True

        # If write protected, check out the file
        elif is_write_protected(file_name):
            from .perforce import perforce_edit
            perforce_edit(file_name, verbose=verbose)

    # Save the file
    if verbose:
        print("Saving {}.".format(file_name))
    save_text_file(file_name, text_lines, line_feed=line_feed, bom=bom)

    # If needed, after the save, mark for add with Perforce
    if do_perforce_add:
        from .perforce import perforce_add
        perforce_add(file_name, verbose=verbose)
    return False

########################################


def environment_root(test_directory, env_name, working_directory=None):
    """
    Check if a directory is relative to an environment variable

    Given a directory, check if it's in the directory of an environment
    variable and if so, return $(NAME)/foo/bar as a path. Otherwise,
    return the path relative to the working_directory or an absolute
    path if working_directory is not set or it's on another drive.

    Args:
        test_directory: Path to test
        env_name: Environment variable to check
        working_directory: Optional root directory
    Returns:
        Tuple with the first is the relative path, the second
            is a boolean if the environment variable is matched
    """

    # Is this environment variable present?
    env_dir = os.environ.get(env_name, None)
    if env_dir:
        try:
            # Is the path relative to the environment variable?
            rel_path = os.path.relpath(test_directory, env_dir)

            # If it doesn't start with dots, it's a match
            if not rel_path.startswith("."):
                return "$(" + env_name + ")" + os.sep + rel_path, True
        except ValueError:
            pass

    # Not a match, try relative or absolute
    if working_directory is None:
        root_path = os.path.abspath(test_directory)
    else:
        try:
            root_path = os.path.relpath(test_directory, working_directory)
        except ValueError:
            # Wrong drive, use absolute path
            root_path = os.path.abspath(test_directory)

    return root_path, False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that locates all Windows SDKs.

@package burger._find_win_sdks

@var burger._find_win_sdks._WIN5_KEYS
List of Windows 5 SDK registry keys

@var burger._find_win_sdks._WIN6_7_KEYS
List of Windows 6-7 SDK registry keys

@var burger._find_win_sdks._WIN6_7SDK_CPUS
List of supported CPUs for Windows 6-7 SDKs

@var burger._find_win_sdks._WIN8_KEYS
List of Windows 8 SDK registry keys

@var burger._find_win_sdks._WIN8SDK_CPUS
List of supported CPUs for Windows 8 SDKs

@var burger._find_win_sdks._WIN8SDK_HEADER_FOLDERS
List of header folders

@var burger._find_win_sdks._WIN8SDK_LIB_FOLDERS
List of library folders

@var burger._find_win_sdks._WIN10SDK_CPUS
List of supported CPUs for Windows 10 SDKs

@var burger._find_win_sdks._WIN10SDK_HEADER_FOLDERS
List of header folders

@var burger._find_win_sdks._WIN10SDK_LIB_FOLDERS
List of library folders

@var burger._find_win_sdks._WIN10_EXECS
List of executables

"""

# pylint: disable=useless-object-inheritance
# pylint: disable=too-many-arguments
# pylint: disable=unnecessary-pass
# pylint: disable=too-few-public-methods
# pylint: disable=consider-using-f-string

from __future__ import absolute_import, print_function, unicode_literals

import os.path

from ._vsinstance import WindowsSDKInstance

try:
    from wslwinreg import convert_from_windows_path
except ImportError:
    pass

from .strutils import make_version_tuple

########################################

# List of Windows 5 SDK registry keys
_WIN5_KEYS = (("7.1", "5.0"), ("8.0", "5.1"))

# List of Windows 6-7 SDK registry keys
_WIN6_7_KEYS = ("v6.0A", "v7.0A", "v7.1A")

# List of supported CPUs for Windows 6-7 SDKs
_WIN6_7SDK_CPUS = ("x86", "x64")

# List of Windows 8 SDK registry keys
_WIN8_KEYS = (("KitsRoot", "8.0"), ("KitsRoot81", "8.1"))

# List of supported CPUs for Windows 8 SDKs
_WIN8SDK_CPUS = ("x86", "x64", "arm")

# List of header folders
_WIN8SDK_HEADER_FOLDERS = ("um", "shared", "winrt")

# List of library folders
_WIN8SDK_LIB_FOLDERS = (("win8", "um"), ("winv6.3", "um"))

# List of supported CPUs for Windows 10 SDKs
_WIN10SDK_CPUS = ("x86", "x64", "arm", "arm64")

# List of header folders
_WIN10SDK_HEADER_FOLDERS = ("ucrt", "um", "shared", "winrt", "cppwinrt")

# List of library folders
_WIN10SDK_LIB_FOLDERS = ("ucrt", "um")

# List of executables
_WIN10_EXECS = ("rc.exe", "signtool.exe", "makecat.exe", "midl.exe", "mc.exe")

########################################


def _find_windows5_sdks(installed_roots):
    """
    Scan the system for all copies of the Windows 5 SDKs

    Windows 5 is Windows XP.

    The SDKs are found by looking at the registry key
    Software\\Microsoft\\Microsoft SDKs\\Windows

    Args:
        installed_roots: Open registry key "Software\\Microsoft"

    Returns:
        List of WindowsSDKInstance for every version of the Windows 6 and 7
        SDK found.
    """

    # pylint: disable=too-many-branches
    # pylint: disable=too-many-nested-blocks

    # Nothing found
    result_list = []

    try:
        key_vc7 = installed_roots.open_subkey(
            "VisualStudio\\SxS\\VC7")
    except OSError:
        return result_list

    # Get the master key, if present
    for key, version_number in _WIN5_KEYS:

        # Get the pathname to the Windows 10 kits, if present.
        try:
            windows_5_path = convert_from_windows_path(
                key_vc7.get_value(key)[0])
        except OSError:
            continue

        windows_5_path = windows_5_path.rstrip("\\/")
        # Initialize the path directory
        known_paths = {}

        # Find the header folder
        test_dir = os.path.join(windows_5_path, "PlatformSDK", "Include")
        if os.path.exists(test_dir):
            known_paths["WinSDK.um"] = test_dir

        # Find all the library folders
        # Note, x86 is assumes, x64 is a sub folder
        for cpu in _WIN6_7SDK_CPUS:
            test_dir = os.path.join(windows_5_path, "PlatformSDK",
                "Lib", "" if cpu == "x86" else "AMD64")
            if os.path.exists(test_dir):
                suffix = "_" + cpu
                known_paths["WinSDK.lib" + suffix] = test_dir

        # Find all executables
        for item in _WIN10_EXECS:
            for cpu in _WIN6_7SDK_CPUS:
                # Try the AMD64 path only
                test_dir = os.path.join(
                    windows_5_path, "PlatformSDK",
                    "bin",
                    "" if cpu == "x86" else "win64" +
                    os.path.sep +
                    "AMD64",
                    item)
                if os.path.exists(test_dir):
                    known_paths[item + "_" + cpu] = test_dir

                # The x86 version has other paths for where the
                # tools are located, because, of course they are.
                elif cpu == "x86":

                    # Try without PlatformSDK
                    test_dir = os.path.join(
                        windows_5_path, "bin", item)
                    if os.path.exists(test_dir):
                        known_paths[item + "_" + cpu] = test_dir
                    else:
                        test_dir = os.path.join(
                            os.path.dirname(windows_5_path),
                            "Common7", "Tools", "Bin", item)
                        if os.path.exists(test_dir):
                            known_paths[item + "_" + cpu] = test_dir

        result_list.append(WindowsSDKInstance(
            "Windows {} SDK".format(version_number[0]),
            version_number,
            windows_5_path,
            known_paths
        ))

    return result_list

########################################


def _find_windows6_7_sdks(installed_roots):
    """
    Scan the system for all copies of the Windows 6 and 7 SDKs

    Windows 6 is Vista, Windows 7 is.. well, you know.

    The SDKs are found by looking at the registry key
    Software\\Microsoft\\Microsoft SDKs\\Windows

    Args:
        installed_roots: Open registry key "Software\\Microsoft"

    Returns:
        List of WindowsSDKInstance for every version of the Windows 6 and 7
        SDK found.
    """

    # pylint: disable=too-many-branches

    # Nothing found
    result_list = []

    try:
        key_sdks = installed_roots.open_subkey(
            "Microsoft SDKs\\Windows")
    except OSError:
        return result_list

    # Get the master key, if present
    for key in _WIN6_7_KEYS:

        # Get the pathname to the Windows 10 kits, if present.
        try:
            sub_key = key_sdks.open_subkey(key)
            windows_6_7_path = convert_from_windows_path(
                sub_key.get_value("InstallationFolder")[0])
        except OSError:
            continue

        try:
            version_number = sub_key.get_value("ProductVersion")[0]
        except OSError:
            continue

        # This hack is to get the version number of the Windows v6.0A
        # SDK because it hides it in one of two sub keys

        # Check if the version starts with "v" (Broken v6.0)
        if version_number[0] == "v":
            try:
                # Try to extract it from the VistaClientWin32Tools subkey
                version_number = sub_key.open_subkey(
                    "VistaClientWin32Tools").get_value("ProductVersion")[0]
            except OSError:
                try:
                    # Try again
                    version_number = sub_key.open_subkey(
                        "VistaClientSDKTools").get_value("ProductVersion")[0]
                except OSError:
                    # Give up and just remove the leading "v"
                    version_number = version_number[1:]
                    continue

        # Initialize the path directory
        known_paths = {}

        # Find the header folder
        test_dir = os.path.join(windows_6_7_path, "Include")
        if os.path.exists(test_dir):
            known_paths["WinSDK.um"] = test_dir

        # Find all the library folders
        # Note, x86 is assumes, x64 is a sub folder
        for cpu in _WIN6_7SDK_CPUS:
            test_dir = os.path.join(windows_6_7_path,
                "Lib", cpu if cpu != "x86" else "")
            if os.path.exists(test_dir):
                suffix = "_" + cpu
                known_paths["WinSDK.lib" + suffix] = test_dir

        # Find all executables
        for item in _WIN10_EXECS:
            for cpu in _WIN6_7SDK_CPUS:
                test_dir = os.path.join(windows_6_7_path,
                        "bin", cpu if cpu != "x86" else "", item)
                if os.path.exists(test_dir):
                    known_paths[item + "_" + cpu] = test_dir

        result_list.append(WindowsSDKInstance(
            "Windows {} SDK".format(version_number[0]),
            version_number,
            windows_6_7_path,
            known_paths
        ))

    return result_list

########################################


def _find_windows8_sdks(installed_roots):
    """
    Scan the system for all copies of the Windows 8 SDKs

    The SDKs are found by looking at the registry key
    Software\\Microsoft\\Windows Kits\\Installed Roots\\KitsRoot8
    and KitsRoot81

    Args:
        installed_roots: Open registry key "Software\\Microsoft"

    Returns:
        List of WindowsSDKInstance for every version of the Windows 8
        SDK found.
    """

    # pylint: disable=too-many-branches

    # Nothing found
    result_list = []

    try:
        roots_key = installed_roots.open_subkey(
            "Windows Kits\\Installed Roots")
    except OSError:
        return result_list

    # Get the master key, if present
    for key, version_number in _WIN8_KEYS:

        # Get the pathname to the Windows 8 kits, if present.
        try:
            windows_8_path = convert_from_windows_path(
                roots_key.get_value(key)[0])
        except OSError:
            continue

        # Try to get the full version number
        try:
            hkey = installed_roots.open_subkey(
                "Microsoft SDKs\\Windows")
            version_number = hkey.open_subkey(
                "v{}A".format(version_number)).get_value("ProductVersion")[0]
        except OSError:
            pass

        # Initialize the path directory
        known_paths = {}

        # Find all the header folders
        for item in _WIN8SDK_HEADER_FOLDERS:
            test_dir = os.path.join(windows_8_path,
                "Include", item)
            if os.path.exists(test_dir):
                known_paths["WinSDK." + item] = test_dir

        # Find all the library folders
        for item, item2 in _WIN8SDK_LIB_FOLDERS:
            for cpu in _WIN8SDK_CPUS:
                test_dir = os.path.join(windows_8_path,
                    "Lib", item, item2, cpu)
                if os.path.exists(test_dir):
                    suffix = item2 if item2 != "um" else ""
                    suffix = suffix + "_" + cpu
                    known_paths["WinSDK.lib" + suffix] = test_dir

        # Find all executables
        for item in _WIN10_EXECS:
            for cpu in _WIN8SDK_CPUS:
                test_dir = os.path.join(windows_8_path,
                        "bin", cpu, item)
                if os.path.exists(test_dir):
                    known_paths[item + "_" + cpu] = test_dir

        result_list.append(WindowsSDKInstance(
            "Windows 8 SDK",
            version_number,
            windows_8_path,
            known_paths
        ))

    return result_list

########################################


def _find_windows10_sdks(installed_roots):
    """
    Scan the system for all copies of the Windows 10 SDKs

    The SDKs are found by looking at the registry key
    Software\\Microsoft\\Windows Kits\\Installed Roots\\KitsRoot10

    Args:
        installed_roots: Open registry key "Software\\Microsoft"

    Returns:
        List of WindowsSDKInstance for every version of the Windows 10
        SDK found.
    """

    # pylint: disable=too-many-branches

    # Nothing found
    result_list = []

    # Get the master key, if present
    try:
        roots_key = installed_roots.open_subkey(
            "Windows Kits\\Installed Roots")
    except OSError:
        return result_list

    # Get the pathname to the Windows 10 kits, if present.
    try:
        windows_10_path = convert_from_windows_path(
            roots_key.get_value("KitsRoot10")[0])
    except OSError:
        return result_list

    # Iterate over the version numbers
    for version_number in os.listdir(
            os.path.join(windows_10_path, "Include")):
        # Convert to version tuple
        version_info = make_version_tuple(version_number)

        # Ensure it's a Windows 10 SDK
        if version_info[0] == 10:

            # Initialize the path directory
            known_paths = {}

            # Find all the header folders
            for item in _WIN10SDK_HEADER_FOLDERS:
                test_dir = os.path.join(windows_10_path,
                    "Include", version_number, item)
                if os.path.exists(test_dir):
                    known_paths["WinSDK." + item] = test_dir

            # Find all the library folders
            for item in _WIN10SDK_LIB_FOLDERS:
                for cpu in _WIN10SDK_CPUS:
                    test_dir = os.path.join(windows_10_path,
                        "Lib", version_number, item, cpu)
                    if os.path.exists(test_dir):
                        suffix = item if item != "um" else ""
                        suffix = suffix + "_" + cpu
                        known_paths["WinSDK.lib" + suffix] = test_dir

            # Find all executables
            for item in _WIN10_EXECS:
                for cpu in _WIN10SDK_CPUS:
                    test_dir = os.path.join(windows_10_path,
                        "bin", version_number, cpu, item)
                    if os.path.exists(test_dir):
                        known_paths[item + "_" + cpu] = test_dir
                        continue
                    test_dir = os.path.join(windows_10_path,
                            "bin", cpu, item)
                    if os.path.exists(test_dir):
                        known_paths[item + "_" + cpu] = test_dir

            result_list.append(WindowsSDKInstance(
                "Windows 10 SDK",
                version_number,
                windows_10_path,
                known_paths
            ))

    return result_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that contains windows only functions

They will operate on Windows, Cygwin, MSYS2 and Windows Subsystem for
Linux. On macOS and pure Linux, they will return no data.

@package burger.windowsutils

@var burger.windowsutils._FIND_VISUAL_STUDIOS
find_visual_studios() cache
"""

from .strutils import get_windows_host_type

from ._find_win_sdks import _find_windows5_sdks, _find_windows6_7_sdks, \
    _find_windows8_sdks, _find_windows10_sdks

from ._find_visual_studio import _find_vs2003_2015, _find_vs2017_higher

try:
    from wslwinreg import get_HKLM_32
except ImportError:
    pass

# find_visual_studios() cache
_FIND_VISUAL_STUDIOS = []

########################################


def find_visual_studios(refresh=False):
    """
    Find every Windows SDK from 5.0 and higher.

    This function may take some time if multiple copies of Visual
    Studio are installed on the machine. For speed, the results are
    cached and the cache is used on subsequent calls.

    More info is here @ref md_find__visual__studio

    Note:
        This function will return an empty list on macOS and pure Linux.
        It has been tested on Windows, Cygwin, MSYS2 and Ubuntu Windows
        Subsystem for Linux.
    Args:
        refresh: Force the cache to be reset if True.
    Returns:
        list of WindowsSDKInstance for every SDK found.
    """

    # pylint: disable=global-statement
    global _FIND_VISUAL_STUDIOS

    result_list = _FIND_VISUAL_STUDIOS
    if not result_list or refresh:
        # Self explanatory. :)

        # Nothing found
        result_list = []

        # Only works on Windows hosted platforms
        if get_windows_host_type(True):

            # Get the root registry key
            try:
                installed_roots = get_HKLM_32().open_subkey(
                    "Software\\Microsoft")
            except OSError:
                installed_roots = None

            if installed_roots:
                result_list.extend(_find_vs2003_2015(installed_roots))
                result_list.extend(_find_vs2017_higher(installed_roots))
                result_list.extend(_find_windows5_sdks(installed_roots))
                result_list.extend(_find_windows6_7_sdks(installed_roots))
                result_list.extend(_find_windows8_sdks(installed_roots))
                result_list.extend(_find_windows10_sdks(installed_roots))

        # Update the cache
        _FIND_VISUAL_STUDIOS = result_list
    return result_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that contains functions to help perforce

@package burger.perforce

@var burger.perforce._PERFORCE_PATH
Cached location of p4 from Perforce
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
import subprocess

try:
    from wslwinreg import convert_from_windows_path, convert_to_windows_path
except ImportError:
    pass

from .strutils import is_string, IS_LINUX, get_windows_host_type, \
    get_mac_host_type

from .buildutils import is_exe, find_in_path, _create_header_guard, \
    run_command, _WINDOWS_ENV_PATHS

# Cached location of p4 from Perforce
_PERFORCE_PATH = None

########################################


def where_is_p4(verbose=False, refresh=False, path=None):
    """
    Return the location of the p4 executable

    Look for an environment variable PERFORCE and
    determine if the executable resides there, if
    so, return the string to the path.

    PATH is then searched for p4, and if it's not found,
    None is returned.

    Args:
        verbose: If True, print a message if Perforce was not found
        refresh: If True, reset the cache and force a reload.
        path: Path to Perforce to place in the cache
    Returns:
        A path to the Perforce command line executable or None if not found.
    See Also:
        perforce_edit, perforce_add, where_is_git, is_under_p4_control
    """

    # pylint: disable=too-many-branches

    # pylint: disable=global-statement
    global _PERFORCE_PATH

    # Clear the cache if needed
    if refresh:
        _PERFORCE_PATH = None

    # Set the override, if found
    if path:
        _PERFORCE_PATH = path

    # Is cached?
    if _PERFORCE_PATH:
        return _PERFORCE_PATH

    # Try the environment variable first
    if os.getenv("PERFORCE", None):
        if get_windows_host_type(True):

            # Windows points to the base path
            p4path = os.path.expandvars("${PERFORCE}\\p4.exe")
            p4path = convert_from_windows_path(p4path)
        else:
            # Just append the exec name
            p4path = os.path.expandvars("${PERFORCE}/p4")

        # Valid?
        if is_exe(p4path):
            _PERFORCE_PATH = p4path
            return p4path

    # Scan the PATH for the exec
    p4path = find_in_path("p4", executable=True)
    if p4path:
        _PERFORCE_PATH = p4path
        return p4path

    # List of the usual suspects
    full_paths = []

    # Check if it's installed but not in the path
    if get_windows_host_type(True):

        # Try the "ProgramFiles" folders
        for item in _WINDOWS_ENV_PATHS:
            if os.getenv(item, None):
                p4path = os.path.expandvars(
                    "${" + item + "}\\perforce\\p4.exe")
                p4path = convert_from_windows_path(p4path)
                full_paths.append(p4path)

    elif get_mac_host_type():

        # Installed here via brew
        full_paths.append("/opt/local/bin/p4")

    if IS_LINUX:
        # Posix / Linux
        full_paths.append("/usr/bin/p4")

    # Scan the list of known locations
    for p4path in full_paths:
        if is_exe(p4path):
            # Finally found it!
            _PERFORCE_PATH = p4path
            return p4path

    # Oh, dear.
    if verbose:
        print("Perforce \"p4\" not found!")
        if get_mac_host_type():
            print("Use brew or macports for the command line version")

    # Can't find it
    return None

########################################


def is_under_p4_control(working_directory):
    """
    Test if the directory is under Perforce source control.

    First test if p4 is installed by calling where_is_p4(). Then
    use the p4 tool to query if the working directory is under Perforce
    source control.

    Note:
        On folders that are not under Perforce control, p4 may take as
        much as 15 seconds to return a result, so use this call with caution.
    Args:
        working_directory: Directory to test.
    Returns:
        True if the directory is under Perforce control, False if not.
    See Also:
        where_is_p4
    """

    p4path = where_is_p4()
    if p4path:
        results = run_command(
            (p4path, "-s", "where", "..."),
            working_directory, True, True, True)
        p4output = results[1].splitlines()
        for item in p4output:
            if item.startswith("exit: "):
                i = int(item[6:])
                if not i:
                    return True
    return False

########################################


def perforce_command(files, command, verbose=False):
    """
    Given a list of files, send a command to execute on them in perforce

    Pass either a single string or a string list of pathnames
    of files to checkout in perforce using the "p4" command with
    the command name

    Args:
        files: list or string object of file(s) to checkout
        command: string to pass to p4 such as "edit" or "add"
        verbose: If True, print the command line and warnings

    Returns:
        Zero if no error, non-zero on error
    See Also:
        where_is_p4
    """

    # Get the p4 executable
    perforce_path = where_is_p4(verbose=verbose)

    # Not found?
    if perforce_path is None:
        return 10

    # Encapsulate the single string entry
    if is_string(files):
        file_list = (files,)
    else:
        file_list = files

    # Generate the command line and call
    error = 0
    for item in file_list:
        item = os.path.abspath(item)
        # If p4.exe, it's windows. Use a windows pathname
        if not perforce_path.endswith("p4"):
            item = convert_to_windows_path(item)

        cmd = [perforce_path, command, item]
        if verbose:
            print(" ".join(cmd))
        error = subprocess.call(cmd)
        if error:
            break
    return error

########################################


def perforce_edit(files, verbose=False):
    """
    Given a list of files, checkout (Edit) them in perforce

    Pass either a single string or a string list of pathnames
    of files to checkout in perforce using the "p4 edit" command

    Args:
        files: list or string object of file(s) to checkout
        verbose: If True, print the command line and warnings

    Returns:
        Zero if no error, non-zero on error
    See Also:
        where_is_p4
    """

    # Perform the edit command
    return perforce_command(files, "edit", verbose=verbose)

########################################


def perforce_add(files, verbose=False):
    """
    Given a list of files, add them in perforce

    Pass either a single string or a string list of pathnames
    of files to checkout in perforce using the "p4 add" command

    Args:
        files: list or string object of file(s) to add
        verbose: If True, print the command line and warnings

    Returns:
        Zero if no error, non-zero on error
    See Also:
        where_is_p4
    """

    # Perform the edit command
    return perforce_command(files, "add", verbose=verbose)

########################################


def perforce_opened(files=None, verbose=False):
    """
    Get the list of opened files in Perforce.

    Check perforce if any files are opened and if so,
    return the list of files in Perforce format that
    are currently opened.

    Args:
        files: List of files or directories to check, None for all.
        verbose: If True, print the command line and warnings.

    Returns:
        List of opened files, can be empty if no files are opened.
    See Also:
        where_is_p4
    """

    # Get the p4 executable
    perforce_path = where_is_p4(verbose=verbose)

    # Not found?
    if perforce_path is None:
        return []

    cmd = [perforce_path, "opened"]

    # Add list of file(s) or directories to check
    if files:
        if is_string(files):
            cmd.append(files)
        else:
            cmd.extend(files)

    # Issue the command
    result = run_command(
        cmd,
        capture_stdout=True,
        capture_stderr=True,
        quiet=not verbose)

    # Was there an error?
    if result[2] != "":
        # Print the error on verbose output
        if verbose:
            print(result[2])
        return []
    # Perforce uses "#" as a delimiter from the filename
    # to the file version.
    return [x.split("#")[0] for x in result[1].splitlines()]

########################################


def make_version_header(working_dir, outputfilename, verbose=False):
    """
    Create a C header with the perforce version.

    This function assumes version control is with perforce!

    Get the last change list and create a header
    with this information (Only modify the output file if
    the contents have changed)

    C++ defines are declared for P4_CHANGELIST, P4_CHANGEDATE, P4_CHANGETIME,
    P4_CLIENT, and P4_USER

    Args:
        working_dir: string with the path of the folder to obtain the perforce
            version for
        outputfilename: string with the path of the generated header
        verbose: Print perforce commands and other informational messages

    Returns:
        Zero if no error, non-zero on error
    """

    # pylint: disable=too-many-branches
    # Check if perforce is installed
    p4exe = where_is_p4()
    if p4exe is None:
        return 10

    # Create the header guard by taking the filename,
    # converting to upper case and replacing spaces and
    # periods with underbars.
    headerguard = _create_header_guard(outputfilename)

    # Get the last change list
    # Parse "Change 3361 on 2012/05/15 13:20:12 by burgerbecky@burgeroctocore
    # 'Made a p4 change'"
    # -m 1 / Limit to one entry
    # -t / Display the time
    # -l / Print out the entire changelist description

    cmd = (p4exe, "changes", "-m", "1", "-t", "-l", "...#have")
    if verbose:
        print(" ".join(cmd))
    error, tempdata = run_command(cmd, working_dir, capture_stdout=True)[:2]
    if error != 0:
        return error

    # Parse out the output of the p4 changes command
    p4changes = tempdata.strip().split(" ")

    # Get the p4 client
    # Parse "P4CLIENT=burgeroctocore (config)"

    cmd = (p4exe, "set", "P4CLIENT")
    if verbose:
        print(" ".join(cmd))
    error, tempdata = run_command(cmd, working_dir, capture_stdout=True)[:2]
    if error != 0:
        return error

    # Parse out the P4CLIENT query
    p4clients = tempdata.strip().split(" ", 1)[0].split("=")

    # Get the p4 user
    # Parse "P4USER=burgerbecky (config)"

    cmd = (p4exe, "set", "P4USER")
    if verbose:
        print(" ".join(cmd))
    error, tempdata = run_command(cmd, working_dir, capture_stdout=True)[:2]
    if error != 0:
        return error

    # Parse out the P4USER query
    p4users = tempdata.strip().split(" ", 1)[0].split("=")

    # Write out the header

    output = [
        "/***************************************",
        "",
        "\tThis file was generated by a call to",
        "\tburger.perforce.make_version_header() from",
        "\tthe burger python package",
        "",
        "***************************************/",
        "",
        "#ifndef " + headerguard,
        "#define " + headerguard,
        ""]

    if len(p4changes) > 4:
        output.append("#define P4_CHANGELIST " + p4changes[1])
        output.append("#define P4_CHANGEDATE \"" + p4changes[3] + "\"")
        output.append("#define P4_CHANGETIME \"" + p4changes[4] + "\"")

    if len(p4clients) > 1:
        output.append("#define P4_CLIENT \"" + p4clients[1] + "\"")

    if len(p4users) > 1:
        output.append("#define P4_USER \"" + p4users[1] + "\"")

    output.extend(["", "#endif"])

    # Check if the data is different than what's already stored on
    # the drive
    # pylint: disable=import-outside-toplevel
    from .fileutils import compare_file_to_string, save_text_file
    if compare_file_to_string(outputfilename, output) is False:
        if verbose:
            print("Writing " + outputfilename)
        try:
            save_text_file(outputfilename, output)
        except IOError as error:
            print(error)
            return 2
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that contains executable locators

Functions that will scan for executables so it's not necessary
for applications to either be on the PATH. They favor environement
variables, and then system registry before searching the path
and then looking at the usual hard coded locations

@package burger.locators

@var burger.locators._CODEBLOCKS_PATH
Cached location of CodeBlocks

@var burger.locators._WATCOM_PATH
Cached location of Watcom

@var burger.locators._DOXYGEN_PATH
Cached location of doxygen

@var burger.locators._PDFLATEX_PATH
Cached location of pdflatex

@var burger.locators._MAKEINDEX_PATH
Cached location of makeindex

@var burger.locators._VS_VARIANTS
Visual Studio variants in the order of search

@var burger.locators._VS_TABLE
Internal table of Visual Studio environment variables
"""

# pylint: disable=consider-using-f-string

from __future__ import absolute_import, print_function, unicode_literals

import os

try:
    from wslwinreg import convert_from_windows_path, OpenKey, CloseKey, \
        KEY_READ, HKEY_CURRENT_USER, KEY_WOW64_32KEY, QueryValueEx, \
        HKEY_LOCAL_MACHINE
except ImportError:
    pass

from .strutils import get_mac_host_type, get_windows_host_type, \
    IS_LINUX, convert_to_windows_slashes

from .buildutils import is_exe, find_in_path, _WINDOWS_ENV_PATHS

from .windowsutils import find_visual_studios

# Cached location of CodeBlocks
_CODEBLOCKS_PATH = None

# Cached location of Watcom
_WATCOM_PATH = None

# Cached location of doxygen
_DOXYGEN_PATH = None

# Cached location of pdflatex
_PDFLATEX_PATH = None

# Cached location of makeindex
_MAKEINDEX_PATH = None

# Visual Studio variants in the order of search
_VS_VARIANTS = (
    "Enterprise",
    "Professional",
    "Community"
)

# For each version of Visual Studio, set the default environment variable
# and path that the specific version of Visual Studio resides
_VS_TABLE = {
    2003: ("VS71COMNTOOLS", "Microsoft Visual Studio .NET 2003", 7),
    2005: ("VS80COMNTOOLS", "Microsoft Visual Studio 8", 8),
    2008: ("VS90COMNTOOLS", "Microsoft Visual Studio 9.0", 9),
    2010: ("VS100COMNTOOLS", "Microsoft Visual Studio 10.0", 10),
    2012: ("VS110COMNTOOLS", "Microsoft Visual Studio 11.0", 11),
    2013: ("VS120COMNTOOLS", "Microsoft Visual Studio 12.0", 12),
    2015: ("VS140COMNTOOLS", "Microsoft Visual Studio 14.0", 14),
    2017: ("VS150COMNTOOLS", "Microsoft Visual Studio\\2017\\xxx", 15),
    2019: ("VS160COMNTOOLS", "Microsoft Visual Studio\\2019\\xxx", 16),
    2022: ("VS170COMNTOOLS", "Microsoft Visual Studio\\2022\\xxx", 17)
}

########################################


def _get_codeblocks_registry_path():
    """
    Locate codeblocks path using the Window Registry

    Private function only called on Windows hosts to query the registry for the
    location of Codeblocks.

    Check HKCU\\SOFTWARE\\CodeBlocks\\Path for the base directory of CodeBlocks
    on Windows.

    Returns:
        None or a string with the path to CodeBlocks
    """

    # Open the registry key
    try:
        key = OpenKey(HKEY_CURRENT_USER, "SOFTWARE\\CodeBlocks",
                      0, KEY_READ | KEY_WOW64_32KEY)
    except FileNotFoundError:
        return None

    # If a key was found, get the path
    codeblocks_path, _ = QueryValueEx(key, "Path")
    CloseKey(key)

    # Convert to path
    codeblocks_path = codeblocks_path + "\\codeblocks.exe"

    # Convert to Linux/Windows/Cygwin
    return convert_from_windows_path(codeblocks_path)


########################################


def where_is_codeblocks(verbose=False, refresh=False, path=None):
    """
    Return the location of CodeBlocks's executable.

    Look for an environment variable CODEBLOCKS and determine if the executable
    resides there, if so, return the string to the path

    If running on a MacOSX client, look in the Applications folder for a copy of
    CodeBlocks.app and return the pathname to the copy of CodeBlocks that
    resides within

    PATH is then searched for CodeBlocks, and if it's not found, None is
    returned.

    Args:
        verbose: If True, print a message if CodeBlocks was not found
        refresh: If True, reset the cache and force a reload.
        path: Path to CodeBlocks to place in the cache

    Returns:
        A path to the CodeBlocks command line executable or None if not found.

    """

    # pylint: disable=too-many-branches

    # pylint: disable=global-statement
    global _CODEBLOCKS_PATH

    # Clear the cache if needed
    if refresh:
        _CODEBLOCKS_PATH = None

    # Set the override, if found
    if path:
        _CODEBLOCKS_PATH = path

    # Is cached?
    if _CODEBLOCKS_PATH:
        return _CODEBLOCKS_PATH

    # Try the environment variable first
    codeblocks_env = os.getenv("CODEBLOCKS", None)
    if codeblocks_env:
        if get_windows_host_type(True):

            # Windows points to the base path
            codeblocks_path = convert_from_windows_path(
                codeblocks_env + "\\codeblocks.exe")
        else:
            # Just append the exec name
            codeblocks_path = os.path.expandvars("${CODEBLOCKS}/CodeBlocks")

        # Valid?
        if is_exe(codeblocks_path):
            _CODEBLOCKS_PATH = codeblocks_path
            return codeblocks_path

    # Scan the PATH for the exec
    codeblocks_path = find_in_path("CodeBlocks", executable=True)
    if codeblocks_path:
        _CODEBLOCKS_PATH = codeblocks_path
        return codeblocks_path

    # List of the usual suspects
    full_paths = []

    # Check if it's installed but not in the path
    if get_windows_host_type(True):

        # Check the registry
        codeblocks_path = _get_codeblocks_registry_path()
        if codeblocks_path:
            full_paths.append(codeblocks_path)

        # Try the "ProgramFiles" folders
        # Note: May not be visible in WSL
        for item in _WINDOWS_ENV_PATHS:
            if os.getenv(item, None):
                codeblocks_path = os.path.expandvars(
                    "${" + item + "}\\CodeBlocks\\codeblocks.exe")
                codeblocks_path = convert_from_windows_path(codeblocks_path)
                full_paths.append(codeblocks_path)

    elif get_mac_host_type():

        # MacOSX has it hidden in the application
        full_paths.append(
            "/Applications/CodeBlocks.app/Contents/MacOS/CodeBlocks")
        full_paths.append("/opt/local/bin/CodeBlocks")

    if IS_LINUX:
        # Posix / Linux
        full_paths.append("/usr/bin/codeblocks")
        full_paths.append("/usr/bin/CodeBlocks")

    # Scan the list of known locations
    for codeblocks_path in full_paths:
        if is_exe(codeblocks_path):
            # Finally found it!
            _CODEBLOCKS_PATH = codeblocks_path
            return codeblocks_path

    # Oh, dear.
    if verbose:
        print("CodeBlocks not found!")
        if get_mac_host_type():
            print("Install the desktop application in the Applications folder")

    # Can't find it
    return None

########################################


def where_is_watcom(command=None, verbose=False, refresh=False, path=None):
    """
    Return the location of Watcom's executables.

    Look for an environment variable WATCOM and determine if the executable
    resides there, if so, return the string to the path

    In Windows, the boot drive is checked for a WATCOM folder and if found,
    that folder name is returned. If all checks failed, None is returned.

    Args:
        command: Watcom program to find.
        verbose: If True, print a message if watcom was not found
        refresh: If True, reset the cache and force a reload.
        path: Path to watcom to place in the cache

    Returns:
        A path to the Watcom folder or None if not found.
    """

    # pylint: disable=too-many-branches

    # pylint: disable=global-statement
    global _WATCOM_PATH

    # Clear the cache if needed
    if refresh:
        _WATCOM_PATH = None

    # Set the override, if found
    if path:
        _WATCOM_PATH = path

    # Windows .exe
    if get_windows_host_type(True):
        exe_folder = "binnt"
        suffix = ".exe"

    # Watcom is not available on macOS yet
    elif get_mac_host_type():
        return None

    # Linux
    else:
        exe_folder = "binl"
        suffix = ""

    # Append the system specific suffix
    if command:
        fake_command = command + suffix
    else:
        fake_command = "wcc386" + suffix

    # Is cached?
    if _WATCOM_PATH:
        if command:
            return os.path.join(_WATCOM_PATH, exe_folder, fake_command)
        return _WATCOM_PATH

    # Try the environment variable first
    watcom_path = os.getenv("WATCOM", None)
    if watcom_path:
        # Valid?
        if get_windows_host_type(True):
            watcom_path = convert_from_windows_path(watcom_path)
        full_path = os.path.join(watcom_path, exe_folder, fake_command)
        if is_exe(full_path):
            _WATCOM_PATH = watcom_path
            return full_path if command else watcom_path

    # List of the usual suspects
    full_paths = []
    if get_windows_host_type(True):
        # Watcom defaults to the root folder
        home_drive = os.getenv("HOMEDRIVE", "C:")
        watcom_path = convert_from_windows_path(home_drive + "\\WATCOM")
        full_paths.append(watcom_path)

        # Try the "ProgramFiles" folders
        for item in _WINDOWS_ENV_PATHS:
            if os.getenv(item, None):
                watcom_path = os.path.expandvars("${" + item + "}\\watcom")
                watcom_path = convert_from_windows_path(watcom_path)
                full_paths.append(watcom_path)

    if IS_LINUX:
        # Posix / Linux
        full_paths.append("/usr/bin/watcom")

    # Scan the list of known locations
    for watcom_path in full_paths:
        # Valid?
        full_path = os.path.join(watcom_path, exe_folder, fake_command)
        if is_exe(os.path.join(watcom_path, full_path)):
            # Finally found it!
            _WATCOM_PATH = watcom_path
            return full_path if command else watcom_path

    # Oh, dear.
    if verbose:
        print("Watcom was not found!")

    # Can't find it
    return None

########################################


def _get_doxygen_registry_path():
    """
    Locate doxygen path using the Window Registry

    Private function only called on Windows hosts to query the registry for the
    location of Doxygen.

    Check HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\
        doxygen_is1 for the base directory of Doxygen on Windows.

    Returns:
        None or a string with the path to CodeBlocks
    """

    # Open the registry key
    try:
        key = OpenKey(HKEY_LOCAL_MACHINE,
        "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\doxygen_is1",
            0, KEY_READ)
    except FileNotFoundError:
        return None

    # If a key was found, get the path
    doxygen_path, _ = QueryValueEx(key, "InstallLocation")
    CloseKey(key)

    # Convert to path
    doxygen_path = convert_to_windows_slashes(
        doxygen_path, True) + "bin\\doxygen.exe"

    # Convert to Linux/Windows/Cygwin
    return convert_from_windows_path(doxygen_path)

########################################


def where_is_doxygen(verbose=False, refresh=False, path=None):
    """
    Return the location of Doxygen's executable

    Look for an environment variable DOXYGEN and determine if the executable
    resides there. Otherwise traverse the path to see if doxygen is present.
    If that fails, look in the registry, or the usual locations like
    "Program Files\\doxygen" on Windows.

    If running on a MacOSX client, look in the Applications folder for a copy of
    Doxygen.app and return the pathname to the copy of doxygen that resides
    within.

    Args:
        verbose: If True, print a message if doxygen was not found
        refresh: If True, reset the cache and force a reload.
        path: Path to doxygen to place in the cache

    Returns:
        A path to the Doxygen command line executable or None if not found.

    """

    # pylint: disable=global-statement
    # pylint: disable=too-many-branches

    global _DOXYGEN_PATH

    # Clear the cache if needed
    if refresh:
        _DOXYGEN_PATH = None

    # Set the override, if found
    if path:
        _DOXYGEN_PATH = path

    # Is cached?
    if _DOXYGEN_PATH:
        return _DOXYGEN_PATH

    # Try the environment variable first
    if os.getenv("DOXYGEN", None):
        if get_windows_host_type(True):

            # Windows points to the base path
            doxygenpath = os.path.expandvars("${DOXYGEN}\\bin\\doxygen.exe")
            doxygenpath = convert_from_windows_path(doxygenpath)
        else:
            # Just append the exec name
            doxygenpath = os.path.expandvars("${DOXYGEN}/doxygen")

        # Valid?
        if is_exe(doxygenpath):
            _DOXYGEN_PATH = doxygenpath
            return doxygenpath

    # Scan the PATH for the exec
    doxygenpath = find_in_path("doxygen", executable=True)
    if doxygenpath:
        _DOXYGEN_PATH = doxygenpath
        return doxygenpath

    # List of the usual suspects
    full_paths = []

    # Check if it's installed but not in the path
    if get_windows_host_type(True):

        # Check the registry
        doxygen_path = _get_doxygen_registry_path()
        if doxygen_path:
            full_paths.append(doxygen_path)

        # Try the "ProgramFiles" folders
        for item in _WINDOWS_ENV_PATHS:
            if os.getenv(item, None):
                doxygen_path = os.path.expandvars(
                    "${" + item + "}\\doxygen\\bin\\doxygen.exe")
                doxygenpath = convert_from_windows_path(doxygen_path)
                full_paths.append(doxygen_path)

    elif get_mac_host_type():

        # MacOSX has it hidden in the application
        full_paths.append(
            "/Applications/Doxygen.app/Contents/Resources/doxygen")
        full_paths.append("/opt/local/bin/doxygen")

    if IS_LINUX:
        # Posix / Linux
        full_paths.append("/usr/bin/doxygen")

    # Scan the list of known locations
    for doxygen_path in full_paths:
        if is_exe(doxygen_path):
            # Finally found it!
            _DOXYGEN_PATH = doxygen_path
            return doxygen_path

    # Oh, dear.
    if verbose:
        print("Doxygen not found!")
        if get_mac_host_type():
            print(
                "Install the desktop application in the Applications folder "
                "or use brew or macports for the command line version")

    # Can't find it
    return None

########################################


def where_is_pdflatex(verbose=False, refresh=False, path=None):
    """
    Return the location of the pdflatex executable

    Args:
        verbose: If True, print a message if pdflatex was not found
        refresh: If True, reset the cache and force a reload.
        path: Path to pdflatex to place in the cache

    Returns:
        A path to the pdflatex command line executable or None if not found.

    """

    # pylint: disable=global-statement

    global _PDFLATEX_PATH

    # Clear the cache if needed
    if refresh:
        _PDFLATEX_PATH = None

    # Set the override, if found
    if path:
        _PDFLATEX_PATH = path

    # Is cached?
    if _PDFLATEX_PATH:
        return _PDFLATEX_PATH

    # Scan the PATH for the exec
    pdflatex_path = find_in_path("pdflatex", executable=True)
    if pdflatex_path:
        _PDFLATEX_PATH = pdflatex_path
        return pdflatex_path

    # List of the usual suspects
    full_paths = []

    # Check if it's installed but not in the path
    if get_windows_host_type(True):

        # Try the "ProgramFiles" folders
        for item in _WINDOWS_ENV_PATHS:
            if os.getenv(item, None):
                pdflatex_path = os.path.expandvars(
                    "${" + item + "}\\texlive\\2024\\bin\\pdflatex.exe")
                pdflatex_path = convert_from_windows_path(pdflatex_path)
                full_paths.append(pdflatex_path)

    elif get_mac_host_type():

        full_paths.append("/opt/local/bin/pdflatex")

    if IS_LINUX:
        # Posix / Linux
        full_paths.append("/usr/bin/pdflatex")

    # Scan the list of known locations
    for pdflatex_path in full_paths:
        if is_exe(pdflatex_path):
            # Finally found it!
            _PDFLATEX_PATH = pdflatex_path
            return pdflatex_path

    # Oh, dear.
    if verbose:
        print("pdflatex not found!")

    # Can't find it
    return None

########################################


def where_is_makeindex(verbose=False, refresh=False, path=None):
    """
    Return the location of the makeindex executable

    Args:
        verbose: If True, print a message if makeindex was not found
        refresh: If True, reset the cache and force a reload.
        path: Path to makeindex to place in the cache

    Returns:
        A path to the makeindex command line executable or None if not found.

    """

    # pylint: disable=global-statement

    global _MAKEINDEX_PATH

    # Clear the cache if needed
    if refresh:
        _MAKEINDEX_PATH = None

    # Set the override, if found
    if path:
        _MAKEINDEX_PATH = path

    # Is cached?
    if _MAKEINDEX_PATH:
        return _MAKEINDEX_PATH

    # Scan the PATH for the exec
    makeindex_path = find_in_path("makeindex", executable=True)
    if makeindex_path:
        _MAKEINDEX_PATH = makeindex_path
        return makeindex_path

    # List of the usual suspects
    full_paths = []

    # Check if it's installed but not in the path
    if get_windows_host_type(True):

        # Try the "ProgramFiles" folders
        for item in _WINDOWS_ENV_PATHS:
            if os.getenv(item, None):
                makeindex_path = os.path.expandvars(
                    "${" + item + "}\\texlive\\2024\\bin\\makeindex.exe")
                makeindex_path = convert_from_windows_path(makeindex_path)
                full_paths.append(makeindex_path)

    elif get_mac_host_type():

        full_paths.append("/opt/local/bin/makeindex")

    if IS_LINUX:
        # Posix / Linux
        full_paths.append("/usr/bin/makeindex")

    # Scan the list of known locations
    for makeindex_path in full_paths:
        if is_exe(makeindex_path):
            # Finally found it!
            _MAKEINDEX_PATH = makeindex_path
            return makeindex_path

    # Oh, dear.
    if verbose:
        print("makeindex not found!")

    # Can't find it
    return None


########################################


def where_is_visual_studio(vs_version, tool_name=None, cpu=None):
    """
    Locate devenv.com for a specific version of Visual Studio.

    Given a specific version by year, check for the appropriate environment
    variable that contains the path to the executable of the IDE

    Note:
        This function will always return None on non-windows hosts.

    Examples:
        >>> burger.where_is_visual_studio(2010)
        "C:\\Program Files (x86)\\Microsoft Visual Studio 10.0"
        "\\Common7\\ide\\devenv.com"

    Args:
        vs_version: Version year as number
        tool_name: Return the path to this tool, None becomes ``devenv.com``
        cpu: String of the cpu type of the tool requested, ``x86``, ``x64``

    Returns:
        Path to devenv.com for the IDE or None.
    """

    # pylint: disable=too-many-branches

    # Test if running on a windows host
    host_type = get_windows_host_type(True)
    if not host_type:
        return None

    # Check if the version is even in the table
    table_item = _VS_TABLE.get(vs_version, None)
    if not table_item:
        return None

    # Tool to locate, use default if not supplied
    if not tool_name:
        tool_name = "devenv.com"

    # Table of cputypes to check
    cputable = []

    # Only check this one
    if cpu:
        cputable.append(cpu)

    else:
        # If already x86, or .com skip extra cpus
        if host_type != "x86" and not tool_name.endswith(".com"):

            # Prioritize the native CPU
            cputable.append(host_type)

            # If ARM, allow x64 for the emulation layer
            if host_type in ("arm", "arm64"):
                cputable.append("x64")

        # Always use x86
        cputable.append("x86")

    # Try the registry first
    for item in find_visual_studios():

        # Is this the version of Visual Studio requested?
        if item.version_info[0] == table_item[2]:

            # Check with CPUs and find a match
            for i in cputable:
                vstudiopath = item.known_paths.get(tool_name + "_" + i, None)
                if vstudiopath:
                    return vstudiopath

    # Try the environment variable next
    vstudio_paths = []
    vstudiopath = os.getenv(table_item[0], default=None)
    if vstudiopath:
        vstudio_paths.append(vstudiopath)

    # Test if this is VS 2017 or higher
    xxx = table_item[1].endswith("xxx")

    # Try the pathname next
    for program_files in _WINDOWS_ENV_PATHS:

        # Generate the proper path to test
        vstudiopath = os.getenv(program_files, None)
        if vstudiopath:

            # Has variants? Call the TVA.
            if xxx:
                for item in _VS_VARIANTS:
                    vstudio_paths.append(
                        vstudiopath +
                        "\\" +
                        table_item[1].replace("xxx", item) +
                        "\\Common7\\Tools\\")
            else:
                vstudio_paths.append(
                    vstudiopath +
                    "\\" +
                    table_item[1] +
                    "\\Common7\\Tools\\")

    for item in vstudio_paths:
        vstudiopath = convert_from_windows_path(item)

        # Locate the launcher
        vstudiopath = os.path.dirname(os.path.abspath(vstudiopath))
        vstudiopath = os.path.join(vstudiopath, "ide", tool_name)
        if os.path.isfile(vstudiopath):
            # Return the path if the file was found
            return vstudiopath

    # Give up
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that contains string manipulation functions

@package burger.strutils

@var burger.strutils.PY2
True if the interpreter is Python 2.x

@var burger.strutils.PY3_OR_HIGHER
True if the interpreter is Python 3.x or higher

@var burger.strutils.PY3_3_OR_HIGHER
True if the interpreter is Python 3.3 or higher

@var burger.strutils.PY3_4_OR_HIGHER
True if the interpreter is Python 3.4 or higher

@var burger.strutils.PY3_5_OR_HIGHER
True if the interpreter is Python 3.5 or higher

@var burger.strutils.PYPY
True if the interpreter is PyPy

@var burger.strutils._WINDOWSSAFESET
Valid characters for windows filenames without quoting

@var burger.strutils._LINUXSAFESET
Valid characters for macOS and Linux files without quoting

@var burger.strutils._RE_COMMA_QUOTES
Regex to match comma and quotes

@var burger.strutils._MAC_HOST_TYPE
Cached result for get_mac_host_type()

@var burger.strutils.IS_LINUX
Running on linux?

@var burger.strutils.IS_MACOSX
Running on macOS X

@var burger.strutils.IS_WINDOWS
Running on Windows

@var burger.strutils.IS_CYGWIN
Running on Cygwin

@var burger.strutils.IS_MSYS
Running on MSYS

@var burger.strutils.IS_WSL
Running on Windows Subsystem for Linux

@var burger.strutils.IS_WINDOWS_HOST
Running on Windows (Including Linux shells on windows)

@var burger.strutils._WINDOWS_HOST_PREFIX
Prefix string for conversion of Windows paths to Linux.

@var burger.strutils._IS_STRING_TEST
Internal type to pass to isinstance() for is_string()

@var burger.strutils.UNICODE
Class for declaring unicode strings

@var burger.strutils.LONG
Class for declaring 64 bit integers
"""

# pylint: disable=consider-using-f-string
# pylint: disable=bad-option-value
# pylint: disable=too-many-lines
# pyright: reportUndefinedVariable=false

from __future__ import absolute_import, print_function, unicode_literals

import sys
import os
import string
import re
import csv
import fnmatch
import platform
from numbers import Number

try:
    from wslwinreg import convert_to_windows_path
except ImportError:
    pass

# True if the interpreter is Python 2.x
PY2 = sys.version_info[0] == 2

# True if the interpreter is Python 3.x or higher
PY3_OR_HIGHER = sys.version_info[0] >= 3

# True if the interpreter is Python 3.3 or higher
PY3_3_OR_HIGHER = sys.version_info >= (3, 3, 0)

# True if the interpreter is Python 3.4 or higher
PY3_4_OR_HIGHER = sys.version_info >= (3, 4, 0)

# True if the interpreter is Python 3.5 or higher
PY3_5_OR_HIGHER = sys.version_info >= (3, 5, 0)

# True if the interpreter is PyPy
PYPY = platform.python_implementation() == "PyPy"

# Valid characters for windows filenames without quoting
_WINDOWSSAFESET = frozenset(string.ascii_letters + string.digits + "_-.:\\")

# Valid characters for macOS and Linux files without quoting
_LINUXSAFESET = frozenset(string.ascii_letters + string.digits + "@%_-+=:,./")

# Regex to match comma and quotes
_RE_COMMA_QUOTES = re.compile(r"\\.|[\"',]", re.DOTALL)

# Cached result for get_mac_host_type()
_MAC_HOST_TYPE = None

# Running on linux?
IS_LINUX = sys.platform.startswith("linux")

# Running on macOS X
IS_MACOSX = sys.platform.startswith("darwin")

# Running on Windows
IS_WINDOWS = sys.platform.startswith("win")

# Running on Cygwin
IS_CYGWIN = sys.platform.startswith("cygwin")

# Running on MSYS
IS_MSYS = sys.platform.startswith("msys")

# Running on Windows Subsystem for Linux
IS_WSL = IS_LINUX and "icrosoft" in platform.platform()

# Running on Windows (Including Linux shells on windows)
IS_WINDOWS_HOST = IS_WINDOWS or IS_CYGWIN or IS_MSYS or IS_WSL

if IS_CYGWIN:
    # Prefix string for conversion of Windows paths to Linux.
    _WINDOWS_HOST_PREFIX = "/cygdrive/"
elif IS_MSYS:
    _WINDOWS_HOST_PREFIX = "/"
elif IS_WSL:
    _WINDOWS_HOST_PREFIX = "/mnt/"
else:
    _WINDOWS_HOST_PREFIX = None

########################################


def unicode_print(input_string):
    """
    Handle printing a unicode string to stdout

    On some platforms, printing a unicode string will trigger
    a UnicodeEncodeError exception. In these cases, handle the
    exception and recode the string to the native string
    encoding.

    Args:
        input_string: A unicode string to print to stdout.
    """

    # Print the string, if no exception, exit
    try:
        print(input_string)
    except UnicodeEncodeError:
        # Ensure it's encoded to utf-8
        encoded = input_string.encode("utf-8")
        if PY2:

            # Python 2.x only accepts this as input
            print(encoded)
        else:

            # Python 3.x and higher will allow remapping to
            # selected character encoding
            print(encoded.decode(sys.stdout.encoding))

########################################


try:
    # Test if basestring exists (Only in Python 2.x)
    # pylint: disable=pointless-statement
    basestring

    # Internal type to pass to isinstance() for is_string()
    _IS_STRING_TEST = basestring

    # Class for declaring unicode strings
    UNICODE = unicode

    # Class for declaring 64 bit integers
    LONG = long

except NameError:
    # Python 3 or later
    _IS_STRING_TEST = (str, bytes)
    UNICODE = str
    LONG = int


def is_string(item):
    """
    Return True if input is a string object

    Test the input if it's either an instance of
    basestring in Python 2.x or (str, bytes) in Python 3.x

    Args:
        item: Object to test
    Returns:
        True if the object is a string instance, False if not.
    """

    return isinstance(item, _IS_STRING_TEST)

########################################


def convert_to_array(input_array):
    """
    Convert a string to a string array (list)

    If the input is None, return an empty list. If
    it's a string, convert the string to a single entry list.
    Otherwise, assume it's an iterable dir, list or tuple of
    strings.

    Args:
        input_array: The object to test
    Returns:
        The input, or a string encapsulated into a single entry list.

    """

    # If empty, return an empty array
    if input_array is None:
        input_array = []
    elif is_string(input_array):
        # Convert a single entry into an array
        input_array = [input_array]
    return input_array

########################################


def norm_paths(working_directory, items):
    """
    Convert an iterable list of pathnames and convert them all into a
    list of normalized absolute pathnames.

    Args:
        working_directory: Base directory for incomplete paths
        items: Iterable list of pathnames
    Returns:
        A list of absolute pathnames
    """

    return [os.path.normpath(os.path.join(working_directory, x)) for x in items]


########################################


def string_to_bool(item):
    """Convert an item to a boolean.

    Strings "true", "t", "on", "yes", and, "y" return True
    "false", "f", "off", "no", and "n" return False

    Non zero numbers or strings that are numbers become True
    Zero and "0" become False.

    Args:
        item: String or integer to convert.

    Returns:
        ``True`` or ``False``

    Exception:
        ValueError on invalid input.
    """

    # pylint: disable=too-many-return-statements
    # None becomes False
    if item is None:
        return False

    # Already a bool?
    if isinstance(item, bool):
        return item

    # Number to bool
    if isinstance(item, Number):
        return bool(item)

    if is_string(item):
        item_lower = item.lower()
        if item_lower in ("true", "t", "on", "yes", "y"):
            return True
        if item_lower in ("false", "f", "off", "no", "n"):
            return False
        try:
            return bool(int(item))
        except ValueError:
            pass
        try:
            return bool(float(item))
        except ValueError:
            pass
    raise ValueError("Can't convert {} to bool".format(item))

########################################


def TrueFalse(item):            # pylint: disable=C0103
    """
    Convert the input into a boolean and return the string "True" or "False"

    If the input was a string of "0" or "False" (Case insensitive comparision),
    this function will return "False". Empty dictionary, string or list
    objects, or the number zero will also return "False"

    Args:
        item: Object to convert to a bool before converting into a string
    Returns:
        The string "True" or "False"
    See Also:
        truefalse, TRUEFALSE
    """

    # Test if it's the string "False"
    if is_string(item):
        if item == "0" or item.upper() == "FALSE":
            return "False"
    if bool(item):
        return "True"
    return "False"

########################################


def truefalse(item):
    """
    Convert the input into a boolean and return the string "true" or "false"

    If the input was a string of "0" or "False" (Case insensitive comparision),
    this function will return "false". Empty dictionary, string or list
    objects, or the number zero will also return "false"

    Args:
        item: Object to convert to a bool before converting into a string
    Returns:
        The string "true" or "false"
    See Also:
        TRUEFALSE, TrueFalse
    """

    # Test if it's the string "False"
    if is_string(item):
        if item == "0" or item.upper() == "FALSE":
            return "false"
    if bool(item):
        return "true"
    return "false"

########################################


def TRUEFALSE(item):            # pylint: disable=C0103
    """
    Convert the input into a boolean and return the string "TRUE" or "FALSE"

    If the input was a string of "0" or "False" (Case insensitive comparision),
    this function will return "FALSE". Empty dictionary, string or list
    objects, or the number zero will also return "FALSE"

    Args:
        item: Object to convert to a bool before converting into a string
    Returns:
        The string "TRUE" or "FALSE"
    See Also:
        truefalse, TrueFalse
    """

    # Test if it's the string "False"
    if is_string(item):
        if item == "0" or item.upper() == "FALSE":
            return "FALSE"
    if bool(item):
        return "TRUE"
    return "FALSE"

########################################


def convert_to_windows_slashes(path_name, force_ending_slash=False):
    """
    Convert a filename from Linux/macOS to Windows format

    Convert all "/" characters into "\" characters

    If force_ending_slash is True, append a "\" if one is not
    present in the final string

    Args:
        path_name: A pathname to be converted to Windows slashes
        force_ending_slash: True if a "\\" character is to be forced at the end
            of the output

    Returns:
        A pathname using Windows type slashes "\\"

    See Also:
        convert_to_linux_slashes

    """

    result = path_name.replace("/", "\\")
    if force_ending_slash and not result.endswith("\\"):
        result = result + "\\"
    return result

########################################


def convert_to_linux_slashes(path_name, force_ending_slash=False):
    """
    Convert a filename from Windows to Linux/macOS format

    Convert all "\" characters into "/" characters

    Args:
        path_name: A string object that text substitution will occur
        force_ending_slash: True if a "/" character is to be forced at the end
            of the output

    Returns:
        A pathname using Linux/BSD type slashes "/"

    See Also:
        convert_to_windows_slashes
    """

    result = path_name.replace("\\", "/")
    if force_ending_slash and not result.endswith("/"):
        result = result + "/"
    return result

########################################


def encapsulate_path_windows(input_path):
    """
    Quote a pathname for use in the Windows system shell

    On Windows platforms, if the path has a space or other
    character that could confuse COMMAND.COM, the string
    will be quoted and double quotes within the string handled
    properly. All slash characters will be replaced with backslash
    characters.

    Args:
        input_path: string with the path to encapsulate using Windows rules
    Returns:
        Original input string if Windows can accept it or input properly quoted
    See Also:
        encapsulate_path
    """

    # Force to Windows slashes
    temp = convert_to_windows_slashes(input_path)

    # If there are any illegal characters, break
    for item in temp:
        if item not in _WINDOWSSAFESET:
            break
    else:
        # No illegal characters in the string
        if not temp:
            return "\"\""
        return temp

    # Since the test failed, quote the string
    return "\"{}\"".format(temp.replace("\"", "\\\""))

########################################


def encapsulate_path_linux(input_path):
    """
    Quote a pathname for use in the linux or BSD system shell

    On Linux platforms, if the path has a space or other
    character that could confuse bash, the string
    will be quoted and double quotes within the string handled
    properly. All backslash characters will be replaced with slash
    characters.

    Args:
        input_path: string with the path to encapsulate using Windows rules
    Returns:
        Original input string if Windows can accept it or input properly quoted
    See Also:
        encapsulate_path
    """

    # Force to linux slashes
    temp = convert_to_linux_slashes(input_path)

    # If there are illegal characters for linux/BSD, break
    for item in temp:
        if item not in _LINUXSAFESET:
            break
    else:
        # String doesn't need quotes
        if not temp:
            return "''"
        return temp
    # Enquote the string for Linux or MacOSX
    return "'{}'".format(temp.replace("'", "'\"'\"'"))

########################################


def encapsulate_path(input_path):
    """
    Quote a pathname for use in the native system shell

    On Windows platforms, if the path has a space or other
    character that could confuse COMMAND.COM, the string
    will be quoted, and for other platforms, it will
    be quoted using rules that work best for BASH.
    This will also quote if the path has a ";" which
    could be used to confuse bash.

    Args:
        input_path: string with the path to encapsulate
    Returns:
        Input string or input properly quoted
    See Also:
        encapsulate_path_windows, encapsulate_path_linux
    """

    # Process for Windows platforms
    if IS_WINDOWS:
        return encapsulate_path_windows(input_path)

    # Force to linux slashes
    return encapsulate_path_linux(input_path)

########################################


def encapsulate_hosted_path(input_path):
    """
    Quote a pathname for use in the windows system shell

    If the platform is hosted on Windows, convert the pathname
    to Windows and then use Windows encapsulation rules.

    Args:
        input_path: string with the path to encapsulate
    Returns:
        Input string or input properly quoted
    See Also:
        encapsulate_path_windows, encapsulate_path_linux, encapsulate_path
    """

    # Process for Windows platforms
    if IS_WINDOWS_HOST:
        return encapsulate_path_windows(convert_to_windows_path(input_path))

    # Force to linux slashes
    return encapsulate_path_linux(input_path)

########################################


def split_comma_with_quotes(comma_string):
    """
    Split comma seperated string while handling quotes

    str.split(",") will split a string into a list but it doesn't
    handle entries that are encased in quotes. This function will
    scan for quote characters and skip over any comma that's encased
    in quotes.

    Examples:
        # Result is ["\"foo,bar\"","foo","bar"]
        lines = burger.strutils.split_comma_with_quotes("\"foo,bar\",foo,bar")

        # Will raise an error due to missing end quote
        willraise = burger.strutils.split_comma_with_quote("\"foo,bar")

    Args:
        comma_string: String of comma seperated strings

    Return:
        List of string fragments for each comma seperated entries

    Raises:
        ValueError

    """

    # Start the parsing at the start of the string
    marker = 0

    # No delimiter found yet
    delimiter = ""
    result = []

    # Get list of matches
    for match in _RE_COMMA_QUOTES.finditer(comma_string):

        # Get the character that matched
        temp = match.group(0)

        # Looking for a comma?
        if delimiter == "":

            # This is a comma that will trigger a split
            if temp == ",":

                # Add in the string that was found
                result.append(comma_string[marker:match.start()])
                # Mark AFTER the comma with +1
                marker = match.start() + 1

            # Is this a quote?
            elif temp in "\"'":
                # Mark a delimiter and stop checking for commas
                delimiter = temp

        # A delimiter is being tracker, has it been hit?
        elif temp == delimiter:
            # Enable splitting on commas
            delimiter = ""

    # If the quote was not matched, throw an exception
    if delimiter:
        raise ValueError("String wasn't properly quoted")

    # Grab the last string chunk
    temp = comma_string[marker:]

    # If it's empty, it's because there was a trailing comma
    if temp:
        result.append(temp)

    return result

########################################


def parse_csv(csv_string):
    """
    Parse a comma seperated string allowing quoted strings

    Given a string of comma seperated entries and handle quotes properly.

    Examples:
        # Result is ["foo,bar","foo","bar"]
        lines = burger.strutils.split_comma_with_quotes("\"foo,bar\",foo,bar")

        # Result is ["foo\"bar","'boo'boo'"]
        lines = burger.strutils.split_comma_with_quotes(
            "\"foo\"\"bar\"","'boo,boo'")

        # Will raise an error due to missing end quote
        willraise = burger.strutils.split_comma_with_quote("\"foo,bar")

    Args:
        csv_string: String of comma seperated entries
    Returns:
        List of entries with whitespace stripped from prefix and suffix
    Raises:
        ValueError
    """

    result = []
    # Seperate by commas
    for item in split_comma_with_quotes(csv_string):

        # Strip the whitespace
        temp = item.strip("\n\r \t")

        # Is there anything to parse?
        if temp:

            # Check if there's a delimiter
            delimiter = temp[0]
            if delimiter in "\"'":

                # If there's a delimiter, properly handle it
                temp = next(
                    csv.reader(
                        [temp],
                        quotechar=str(delimiter),
                        delimiter=str(","),
                        quoting=csv.QUOTE_ALL))[0]
        result.append(temp)
    return result

########################################


def translate_to_regex_match(file_list):
    """
    Translate filename wildcards into regexes.

    Args:
        file_list: List of filename wildcards
    Results:
        List of re.compile().match entries
    """

    # Translate and then return the match function
    return [re.compile(fnmatch.translate(
        x)).match for x in convert_to_array(file_list)]

########################################


def host_machine():
    """
    Return the high level operating system's name

    Return the machine this script is running on, "windows", "macosx",
    "linux" or "unknown"

    Returns:
        The string "windows", "macosx", "linux", or "unknown"

    See Also:
        get_mac_host_type, get_windows_host_type
    """

    # Is it ONLY windows, not hosted on windows?
    if IS_WINDOWS:
        return "windows"

    # Is it macOS X?
    if IS_MACOSX:
        return "macosx"

    if IS_LINUX or IS_CYGWIN or IS_MSYS or IS_WSL:
        # Assume linux or linux clones
        return "linux"

    # Surrender Dorothy
    return "unknown"

########################################


def get_windows_host_type(wsl_allowed=False):
    """
    Return windows host type (32 or 64 bit)

    Return False if the host is not Windows, "x86" if it's a 32 bit host
    and "x64" if it's a 64 bit host, and possibly "arm" if an arm host

    Args:
        wsl_allowed: If True, allow returning a host type if cygwin
            is the shell.
    Returns:
        The string "x64", "x86", "arm", "arm64", "ia64" or False
    See Also:
        get_mac_host_type, host_machine

    """

    # Not windows?
    if not IS_WINDOWS_HOST:
        return False

    # Is Linux under windows allowed?
    if not IS_WINDOWS:
        # Handle the case if host is Cygwin/msys/Linux running on windows
        if not wsl_allowed:
            return False

    # Test the CPU for the type

    machine = platform.machine().lower()
    if machine in ("amd64", "x86_64", "em64t"):
        return "x64"
    if machine in ("arm64", "arm", "ia64"):
        return machine
    return "x86"

########################################


def get_mac_host_type():
    """
    Return Mac OSX host type (PowerPC/Intel)

    Return False if the host is not Mac OSX. "ppc" or "ppc64" if it's a Power
    PC based system, "x86" or "x64" for Intel (Both 32 and 64 bit)

    Returns:
        The string "x86", "x64", "ppc", "ppc64" or False.

    See Also:
        get_windows_host_type, host_machine
    """

    global _MAC_HOST_TYPE       # pylint: disable=W0603
    if _MAC_HOST_TYPE is None:

        # Not macOS? Force returning False
        if not IS_MACOSX:
            cpu = False
        else:
            # Since it's a mac, query the Mac OSX cpu type
            # using the MacOSX python extensions

            cpu = platform.machine()

            # Intel 64 bit
            if cpu == "x86_64":
                cpu = "x64"

            # Special case for Power Macs
            elif cpu in ("PowerPC", "ppc", "Power Macintosh"):

                # Issue the command `machine` and capture the output.
                # If it contains "ppc970" then you're running on a 64 bit
                # version of macOS, despite what platform.machine() says

                # pylint: disable=import-outside-toplevel
                from .buildutils import run_command
                if "ppc970" in run_command("machine", capture_stdout=True)[1]:
                    cpu = "ppc64"
                else:
                    cpu = "ppc"
        _MAC_HOST_TYPE = cpu
    # Return the resolved global
    return _MAC_HOST_TYPE


########################################

def escape_xml_cdata(xml_string):
    """
    Convert escape codes for CDATA XML records.

    According to the XML docs, &, < and > cannot exist in a string
    so they must be replaced with "&amp;", "&lt;" and "&gt;" respectively.

    Args:
        xml_string: String to convert to one compatible with XML CDATA
    Return:
        Original string if no changes, or a new string with escaped characters.
    """

    # For speed, only perform the replace operation on strings that
    # actually need conversion.

    # IMPORTANT! Convert & first, since it will be inserted in
    # operations that follow.
    if "&" in xml_string:
        xml_string = xml_string.replace("&", "&amp;")

    if "<" in xml_string:
        xml_string = xml_string.replace("<", "&lt;")
    if ">" in xml_string:
        xml_string = xml_string.replace(">", "&gt;")
    return xml_string

########################################


def escape_xml_attribute(xml_string):
    """
    Convert escape codes for XML element attribute records.

    According to the XML docs, ", &, < and > cannot exist in a string
    so they must be replaced with "&quot;", &amp;", "&lt;" and "&gt;"
    respectively.

    Tabs and line feeds will be converted to "&#10;" and "&#09;".

    Note:
        https://www.w3.org/TR/REC-xml/#sec-line-ends
    Args:
        xml_string: String to convert to XML attribute strings.
    Return:
        Original string if no changes, or a new string with escaped characters.
    """

    # For speed, only perform the replace operation on strings that
    # actually need conversion.

    # Process &, < and >
    xml_string = escape_xml_cdata(xml_string)

    # Attributes are encased in quotes, escape the quote
    if "\"" in xml_string:
        xml_string = xml_string.replace("\"", "&quot;")

    # Line feeds need to be handled carefully because no one
    # can agree on \r, \n, \r\n, see note URL above
    if "\r\n" in xml_string:
        xml_string = xml_string.replace("\r\n", "&#10;")
    if "\r" in xml_string:
        xml_string = xml_string.replace("\r", "&#10;")
    if "\n" in xml_string:
        xml_string = xml_string.replace("\n", "&#10;")

    # Let's convert tabs
    if "\t" in xml_string:
        xml_string = xml_string.replace("\t", "&#09;")
    return xml_string

########################################


def packed_paths(entries, slashes=None, separator=None,
                 force_ending_slash=False):
    """
    Convert a list of paths and convert to a PATH string.

    Convert ["a","b","c"] to "a;b;c"

    Args:
        entries: list of strings to concatenate
        separator: Character to use to seperate entries, ";" is used for None
        force_ending_slash: Enforce a trailing slash if True
        slashes: None for no conversion "/" or "\\" path separator
    Return:
        String of all entries seperated by ";" or ``separator``
    """

    # Verify the separator
    if not separator:
        separator = ";"

    entries = convert_to_array(entries)

    # Slash functionality?
    if slashes:
        # Windows?
        if slashes == "\\":
            function = convert_to_windows_slashes
        else:
            # Everyone else use linux slashes
            function = convert_to_linux_slashes

        # Don't modify the original list
        temp_entries = [
            function(
                x,
                force_ending_slash=force_ending_slash) for x in entries]
    else:
        temp_entries = entries
    return separator.join(temp_entries)

########################################


def make_version_tuple(version_string):
    """
    Convert numeric version string into an int tuple.

    Given a string like "1.0.5rc" and return a tuple of (1, 0, 5).
    The numbers are seperated by periods and members that start with
    a non number are skipped.

    Args:
        version_string: String for the version number.
    Returns:
        tuple of integers with the version. Can be an empty tuple.
    """

    if not version_string or not is_string(version_string):
        return tuple()

    result = []
    for item in version_string.split("."):
        # If a pure number, convert to an int
        try:
            result.append(int(item))
        except ValueError:
            # None numeric, parse out the numbers
            found = re.findall(r"\d+", item)
            if not found:
                # Not a number? Abort here.
                break
            result.append(int(found[0]))
    return tuple(result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that contains git helper functions

Functions to locate git and issue git commands

@package burger.git

@var burger.git._GIT_PATH
Cached location of git
"""

from __future__ import absolute_import, print_function, unicode_literals

import os

try:
    from wslwinreg import convert_from_windows_path
except ImportError:
    pass

from .strutils import get_windows_host_type, IS_LINUX, get_mac_host_type

from .buildutils import is_exe, _WINDOWS_ENV_PATHS, find_in_path, \
    run_command, _create_header_guard

# Cached location of git
_GIT_PATH = None

########################################


def where_is_git(verbose=False, refresh=False, path=None):
    """
    Return the location of the git executable

    Look for an environment variable GIT and
    determine if the executable resides there, if
    so, return the string to the path.

    PATH is then searched for git, and if it's not found,
    None is returned.

    Args:
        verbose: If True, print a message if git was not found
        refresh: If True, reset the cache and force a reload.
        path: Path to git to place in the cache
    Returns:
        A path to the git command line executable or None if not found.
    See Also:
        where_is_p4, is_under_git_control
    """
    # pylint: disable=too-many-branches

    # pylint: disable=global-statement
    global _GIT_PATH

    # Clear the cache if needed
    if refresh:
        _GIT_PATH = None

    # Set the override, if found
    if path:
        _GIT_PATH = path

    # Is cached?
    if _GIT_PATH:
        return _GIT_PATH

    # Try the environment variable first
    if os.getenv("GIT", None):
        if get_windows_host_type(True):

            # Windows points to the base path
            gitpath = os.path.expandvars("${GIT}\\git.exe")
            gitpath = convert_from_windows_path(gitpath)
            if is_exe(gitpath):
                _GIT_PATH = gitpath
                return gitpath
            # Try a second time using the bin folder
            gitpath = os.path.expandvars("${GIT}\\bin\\git.exe")
            gitpath = convert_from_windows_path(gitpath)
        else:
            # Just append the exec name
            gitpath = os.path.expandvars("${GIT}/git")

        # Valid?
        if is_exe(gitpath):
            _GIT_PATH = gitpath
            return gitpath

    # Scan the PATH for the exec
    gitpath = find_in_path("git", executable=True)
    if gitpath:
        _GIT_PATH = gitpath
        return gitpath

    # List of the usual suspects
    full_paths = []

    # Check if it's installed but not in the path
    if get_windows_host_type(True):

        # Try the "ProgramFiles" folders
        for item in _WINDOWS_ENV_PATHS:
            if os.getenv(item, None):
                gitpath = os.path.expandvars(
                    "${" + item + "}\\git\\bin\\git.exe")
                gitpath = convert_from_windows_path(gitpath)
                full_paths.append(gitpath)

    elif get_mac_host_type():

        # Installed here via brew
        full_paths.append("/opt/local/bin/git")

    if IS_LINUX:
        # Posix / Linux
        full_paths.append("/usr/bin/git")

    # Scan the list of known locations
    for gitpath in full_paths:
        if is_exe(gitpath):
            # Finally found it!
            _GIT_PATH = gitpath
            return gitpath

    # Oh, dear.
    if verbose:
        print("git not found!")
        if get_mac_host_type():
            print("Use brew or macports for the command line version")

    # Can't find it
    return None

########################################


def is_under_git_control(working_directory):
    """
    Test if the directory is under git source control.

    First test if git is installed by calling where_is_git(). Then
    use the git tool to query if the working directory is under git
    source control.

    Args:
        working_directory: Directory to test.
    Returns:
        True if the directory is under git control, False if not.
    See Also:
        where_is_git
    """

    gitpath = where_is_git()
    if gitpath:
        if not run_command(
            (gitpath, "rev-parse"),
                working_directory, True, True, True)[0]:
            return True
    return False

########################################


def _call_git(cmd, working_dir, verbose):
    """
    Call git and return the output.

    If an error occured, return None for the string.

    Args:
        cmd: Tuple of the command line to execute
        working_dir: Directory to set before executing the command
        verbose: True for verbose output.
    Returns:
        Error code integer, returned string from git.
    """

    # Output the command line if requested
    if verbose:
        print(" ".join(cmd))

    # Perform the command
    # If verbose output is enabled, allow git to print the error
    error, tempdata = run_command(
        cmd, working_dir, capture_stdout=True, capture_stderr=not verbose)[:2]

    # If there was an error, discard the output.
    if error:
        return error, None

    # All good
    return error, tempdata.strip()

########################################


def make_git_version_header(working_dir, outputfilename, verbose=False):
    """
    Create a C header with the git version.

    This function assumes version control is with git!

    Get the last change list and tag, and then create a header
    with this information (Only modify the output file if
    the contents have changed)

    C++ defines are declared for GIT_HASH, GIT_CHANGEDATE, GIT_CHANGETIME,
    GIT_TAG_VERSION, and GIT_TAG_VERSION_INFO

    Args:
        working_dir: string with the path of the folder to obtain the git
            version for
        outputfilename: string with the path of the generated header
        verbose: Print git commands and other informational messages

    Returns:
        Zero if no error, non-zero on error
    """

    # pylint: disable=too-many-branches

    # Check if git is installed
    gitexe = where_is_git()
    if gitexe is None:
        return 10

    # Create the header guard by taking the filename,
    # converting to upper case and replacing spaces and
    # periods with underbars.
    headerguard = _create_header_guard(outputfilename)

    # Get the last hash
    error, git_hash = _call_git(
        (gitexe, "rev-parse", "HEAD"),
        working_dir, verbose)

    # The only way there is an error, is if this directory is not controlled
    # by git.
    if error:
        return error

    # Get the current branch
    error, git_branch = _call_git(
        (gitexe, "rev-parse", "--abbrev-ref", "HEAD"),
        working_dir, verbose)

    # Get the git tag
    error, git_tag = _call_git(
        (gitexe, "describe", "--tags", "--abbrev=0"),
        working_dir, verbose)

    # Get the full git tag
    error, git_full_tag = _call_git(
        (gitexe, "describe", "--tags", "--long"),
        working_dir, verbose)

    # Write out the header

    output = [
        "/***************************************",
        "",
        "\tThis file was generated by a call to",
        "\tburger.git.make_git_version_header() from",
        "\tthe burger python package",
        "",
        "***************************************/",
        "",
        "#ifndef " + headerguard,
        "#define " + headerguard,
        ""]

    if git_hash:
        output.append("#define GIT_HASH \"" + git_hash + "\"")

    if git_branch:
        output.append("#define GIT_BRANCH \"" + git_branch + "\"")

    if git_full_tag:
        output.append("#define GIT_FULL_TAG \"" + git_full_tag + "\"")

    if git_tag:
        output.append("#define GIT_TAG \"" + git_tag + "\"")

    output.extend(["", "#endif"])

    # Check if the data is different than what's already stored on
    # the drive
    # pylint: disable=import-outside-toplevel
    from .fileutils import compare_file_to_string, save_text_file
    if compare_file_to_string(outputfilename, output) is False:
        if verbose:
            print("Writing " + outputfilename)
        try:
            save_text_file(outputfilename, output)
        except IOError as error:
            print(error)
            return 2
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that contains xcode helper functions

Functions to locate xcode and issue rez commands

@package burger.xcode

@var burger.xcode._CODEWARRIOR_LOCATIONS
Locations of CodeWarrior 9 or 10
"""

# pylint: disable=consider-using-f-string

from __future__ import absolute_import, print_function, unicode_literals

import os

from .strutils import PY3_4_OR_HIGHER, get_mac_host_type, convert_to_array
from .buildutils import run_command

# Locations of where Codewarrior is installed
_CODEWARRIOR_LOCATIONS = (
    "/Applications/Metrowerks CodeWarrior 10.0",
    "/Applications/Metrowerks CodeWarrior 9.0"
)

########################################


def where_is_xcode(xcode_version=None):
    """
    Locate xcodebuild for a specific version of XCode.

    Given a specific version by version, scan the locations that the IDE would
    be found.

    Example:
        >>> burger.where_is_xcode()
        ("/Developer/usr/bin/xcodebuild", 3)
        >>> burger.where_is_xcode(2093)
        None

    Note:
        This function will always return None on non-macOS hosts.
        Minimum version of XCode is 3.

    Args:
        xcode_version: Version number
    Returns:
        Path to xcodebuild for the XCode version or None.
    """

    # pylint: disable=too-many-branches

    # Test if running on a mac host
    host_type = get_mac_host_type()
    if not host_type:
        return None

    # Only import on macosx hosts
    # pylint: disable=import-outside-toplevel
    import plistlib

    # XCode 5 and higher reside in the app folder
    highest_version = 0
    xcodebuild = None

    # Version 3 and 4 is in /Developer while all
    # others are in /Applications

    dir_list = []
    if xcode_version is None or xcode_version < 5:
        dir_list.append("/Developer/Applications")

    if xcode_version is None or xcode_version > 3:
        dir_list.append("/Applications")

    for base_dir in dir_list:
        # Check if the directory exists first
        if os.path.isdir(base_dir):

            # Scan the applications folder for all apps called "XCode"
            for item in os.listdir(base_dir):

                # Scan only apps whose name starts with xcode
                if not item.lower().startswith("xcode"):
                    continue

                temp_path = base_dir + "/" + item + "/Contents/version.plist"
                try:
                    # pylint: disable=no-member
                    if PY3_4_OR_HIGHER:
                        with open(temp_path, "rb") as filefp:
                            version_dict = plistlib.load(filefp)
                    else:
                        version_dict = plistlib.readPlist(
                            temp_path)

                # Any IO error is acceptable to ignore
                except IOError:
                    continue

                version = version_dict.get("CFBundleShortVersionString", None)
                if not version:
                    continue

                # Check the version for a match
                version = int(version.split(".")[0])

                # XCode 3 is hard coded to Developer
                if version == 3:
                    temp_path = "/Developer/usr/bin/xcodebuild"
                else:
                    temp_path = (
                        "{}/{}/Contents/Developer"
                        "/usr/bin/xcodebuild").format(base_dir, item)

                if not os.path.isfile(temp_path):
                    continue

                if xcode_version:
                    # If scanning for a perfect match, exit if found
                    if version == xcode_version:
                        highest_version = version
                        return (temp_path, version)

                # Scan for the most recent version of XCode
                elif version > highest_version:
                    highest_version = version
                    xcodebuild = (temp_path, version)

    # XCode 3 is hard coded to a specific location
    if (not xcode_version and not highest_version) or xcode_version == 3:

        # On OSX Lion and higher, XCode 3.1.4 is a separate folder
        for item in ("/Xcode3.1.4/usr/bin/xcodebuild",):
            if os.path.isfile(item):
                xcodebuild = (item, 3)
                break

    return xcodebuild


########################################


def find_rez_headers():
    """
    Locate where the resource headers are located

    Of course Apple would not be consistent where the SDKs and headers are
    located, so search for them.

    If successful, a tuple is returned with the folder to pass to ``rez`` as
    the folder that has the *.r headers as the first entry.

    If running on on Xcode 5 or higher, the SDKs is built into the app and
    the ``FlatCarbon`` headers are missing. As a result, the second entry of
    the tuple will be ``True`` to denote that Frameworks should be used instead
    of classic MacOS 9 flat header folders. If classic headers, the second
    entry of the tuple is ``False``

    Returns:
        (Path to the rez headers, True/False) or None
    """

    # Only valid on mac hosts
    if not get_mac_host_type():
        return None

    # Step 1, see if CodeWarrior is installed. If so,
    # stick to the classics

    for item in _CODEWARRIOR_LOCATIONS:

        # Located?
        if os.path.isdir(item):

            # Create the path and exit
            return (os.path.join(
                item,
                "Metrowerks CodeWarrior",
                "MacOS Support",
                "Universal",
                "Interfaces",
                "RIncludes"
            ), False)

    # At this point, look for the includes in Xcode
    xcode_ref = where_is_xcode()
    if not xcode_ref:
        # XCode not found
        return None

    # Get the directory and Xcode version
    xcode_dir = xcode_ref[0]
    # xcode_version = xcode_ref[1]

    # Remove usr/bin/xcodebuild
    xcode_dir = os.path.dirname(
        os.path.dirname(
            os.path.dirname(xcode_dir)))

    # Are we a really old version of Xcode, like
    # 2, 3, or 4?
    if xcode_dir.startswith("/Developer"):

        # Go to the base directory of the SDKs
        xcode_dir = os.path.join(xcode_dir, "SDKs")

        # Since there's no hard link to the default, manually iterate
        # over the possible choices

        # Reverse sort to get the most recent first
        for item in sorted(os.listdir(xcode_dir), reverse=True):

            # Skip files, and use the first one hit
            if os.path.isdir(os.path.join(xcode_dir, item)):

                # Classic Xcode has a flattened folder for headers
                # so classic paths still work
                return (os.path.join(
                    xcode_dir,
                    item,
                    "Developer",
                    "Headers",
                    "FlatCarbon"), False)

        # No sdks installed?
        return None

    # We are a modern version of Xcode, which
    # means the SDKs are inside the app.

    # Append the path, seriously, this is the path
    # No, seriously, what were they smoking?
    return (os.path.join(
        xcode_dir,
        "Platforms",
        "MacOSX.platform",
        "Developer",
        "SDKs",
        # Use the default sdk
        "MacOSX.sdk"), True)

########################################


def build_rez(working_directory, input_file,
              output_file, creator=None, filetype=None,
              frameworks=None):
    """
    Process the Rez file using the rez tool

    Determine where the ``rez`` tool resides, as well as the MacOS rez headers.

    Runs the ``rez`` tool with the header directory properly preset. Will also
    set the creator and filetype if they are passed with 4 character strings.

    Modern versions of XCode allows the invokation of specific frameworks, pass
    them in in the frameworks parameter. It accepts a single string or an
    iterable of strings.

    Note:
        Only runs on a Mac

    Args:
        working_directory
            Directory this script resides in.

        input_file
            Name of the .r file to compile with rez

        output_file
            Name of the output file

        creator
            4 character string for the creator code

        filetype
            4 character string for the filetype

        frameworks
            iterable of frameworks that would be invoked on modern xcode

    Returns:
        Zero on success, non zero on error
    """

    # Only valid on mac hosts
    if not get_mac_host_type():
        return 10

    # Are the resource headers present?
    rez_ref = find_rez_headers()
    if rez_ref is None:
        return 10

    # Get the directory and the style
    rez_dir = rez_ref[0]
    new_style = rez_ref[1]

    # Invoke the rez compiler
    cmd = ["rez"]

    # Classic style?
    if not new_style:

        # Rez supports file creator and type so set them if needed
        if creator:
            cmd.append("-c")
            cmd.append(creator)

        if filetype:
            cmd.append("-t")
            cmd.append(filetype)

        # Location of resource headers
        cmd.append("-i")
        cmd.append(rez_dir)
    else:
        # New style means using frameworks

        # Alert the source file that frameworks are in play
        cmd.append("-d")
        cmd.append("USING_FRAMEWORKS")

        # Location of the sdk
        cmd.append("-is")
        cmd.append(rez_dir)

        # Use these frameworks
        if frameworks:
            for item in convert_to_array(frameworks):
                cmd.append("-F")
                cmd.append(item)

    # Append output name
    cmd.append("-o")
    cmd.append(output_file)

    # Append the input name
    cmd.append(input_file)

    # Run the compiler
    error, _, _ = run_command(cmd, working_directory)

    if not error and new_style:
        # New style Rez generator doesn't set the filetype or creator

        # Do it manually if needed
        if creator or filetype:
            cmd = ["SetFile"]

            if creator:
                cmd.append("-c")
                cmd.append(creator)

            if filetype:
                cmd.append("-t")
                cmd.append(filetype)

            cmd.append(output_file)

            error, _, _ = run_command(cmd, working_directory)

    # Return the error code
    return error



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package containing shared classes for Windows SDK and Visual Studio
instance searches.

@package burger._vsinstance
"""

# pylint: disable=useless-object-inheritance
# pylint: disable=consider-using-f-string

from .strutils import make_version_tuple

########################################


class VisualStudioInstance(object):
    """
    Instance of the location of Visual Studio.

    More info is here @ref md_find__visual__studio

    Attributes:
        name: Name of the Visual Studio instance or SDK
        version_string: String of the version of this SDK
        version_info: Integer tuple of self.version
        path: Root pathname of the object
        known_paths: dict of pathnames for specific items

    See Also:
        WindowsSDKInstance
    """

    # pylint: disable=too-many-arguments
    # pylint: disable=too-few-public-methods

    def __init__(self, name, ver, path, known_paths):
        """
        Initialize a VisualStudioInstance

        Args:
            name: Name of the SDK or Visual Studio instance
            ver: Version string.
            path: Base path to the SDK or Visual Studio instance.
            known_paths: dir of paths for specific executables and folders.
        """

        self.name = name
        self.version_string = ver
        self.version_info = make_version_tuple(ver)
        self.path = path.rstrip("\\/")
        self.known_paths = known_paths

    def __repr__(self):
        """
        Print the name of the class instance.
        """
        return "<{} {} at {}>".format(
            type(self).__name__, self.version_string, self.path)

    def __str__(self):
        """
        Print the name of instance.
        """
        return self.name

########################################


class WindowsSDKInstance(VisualStudioInstance):
    """
    Instance of paths for a Windows SDK.

    More info is here @ref md_find__visual__studio

    See Also:
        VisualStudioInstance
    """

    # pylint: disable=unnecessary-pass
    # pylint: disable=too-few-public-methods
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that contains class member validators

@package burger.validators
"""

# pylint: disable=no-name-in-module,too-few-public-methods
# pylint: disable=consider-using-f-string
# pylint: disable=deprecated-class

from numbers import Number

# Try the Python 3 import
try:
    from collections.abc import Iterable
except ImportError:
    # Try the Python 2 import
    from collections import Iterable

from .strutils import string_to_bool, is_string, UNICODE as unicode, \
    LONG as long

# Note: (object) is required for python 2.7 compatiblity
# pylint: disable=useless-object-inheritance


class Property(object):
    """
    Base Class to create enforced types

    Attributes:
        _name: The real name of the class instance
    """

    def __init__(self, name):
        """Initialize to default
        Args:
            name: Name of the instance storage index
        """

        self._name = name

    def __get__(self, instance, owner=None):
        """Return value

        Args:
            instance: Reference to object containing data
            owner: Not used

        Returns:
            None, or verified data
        """

        return instance.__dict__.get(self._name, None)

    def __set__(self, instance, value):
        """Set the string value

        Args:
            instance: Reference to object containing data
            value: None or value the can be converted to bool
        """

        instance.__dict__[self._name] = value

########################################


class BooleanProperty(Property):
    """Class to enforce bool in member variable

See Also:
    strutils.string_to_bool

Example:
"Inherit from (object) for Python 2.7"
>>> class foo(object):
    "Init to false"
    x = BooleanProperty(False)
    "Init to true"
    y = BooleanProperty(True)
    "Init to None"
    z = BooleanProperty()

"Create the class"
f = foo()

"Print True"
print(f.x)

"Print False"
f.x = False
print(f.x)

"f.x is set to bool False with string"
f.x = "False"
print(f.x)

"Exception on bad write"
f.x = "not boolean"
Traceback (most recent call last):
    ...
ValueError: Not boolean value
    """

    def __set__(self, instance, value):
        """Set the boolean value

        Args:
            instance: Reference to object containing data
            value: None or value the can be converted to bool

        Exception:
            ValueError on invalid input.

        See Also:
            strutils.string_to_bool
        """

        if value is not None:
            # Convert to bool
            value = string_to_bool(value)

        # Boolean value
        instance.__dict__[self._name] = value

########################################


class IntegerProperty(Property):
    """Class to enforce 64 bit integer in variable

Example:
"Inherit from (object) for Python 2.7"
>>> class foo(object):
    "Init to 1"
    x = IntegerProperty(1.0)
    "Init to 55"
    y = IntegerProperty("55")
    "Init to None"
    z = IntegerProperty()

"Create the class"
f = foo()

"Print 1"
print(f.x)

"Print 0"
f.x = False
print(f.x)

"f.x is set to 99 with string"
f.x = "99.00"
print(f.x)

"Exception on bad write"
f.x = "not boolean"
Traceback (most recent call last):
    ...
ValueError: Not integer value
    """

    def __set__(self, instance, value):
        """Set the integer value

        Args:
            instance: Reference to object containing data
            value: None or value the can be converted to bool

        Exception:
            ValueError on invalid input.

        See Also:
            strutils.string_to_bool

        """

        if value is not None:
            # Bool is a special case (0,1)
            if isinstance(value, bool):
                value = int(value)
            else:
                # Not a number?
                if not isinstance(value, Number):

                    # Convert string to integer
                    if is_string(value):
                        # Convert to integer
                        try:
                            value = long(value, 0)
                        # Try again as a float
                        except ValueError:
                            value = float(value)
                    else:
                        raise ValueError(
                            "Value \"{}\" is not a number".format(value))

                # value is a number, bounds check it
                if value < -0x8000000000000000 or value > 0x7fffffffffffffff:
                    raise ValueError(
                        "Value \"{}\" must fit in signed 64 bits".format(value))

                # Ensure it's an int()
                value = long(value)

        # Integer value
        instance.__dict__[self._name] = value

########################################


class StringProperty(Property):
    """Class to enforce string in member variable

Example:
"Inherit from (object) for Python 2.7"
>>> class foo(object):
    "Init to \"foo\""
    x = StringProperty("foo")
    "Init to None"
    y = StringProperty()

"Create the class"
f = foo()

"Print foo"
print(f.x)

"Print False"
f.x = "False"
print(f.x)

"Print True"
f.x = True
print(f.x)
    """

    def __set__(self, instance, value):
        """Set the string value
        Args:
            instance: Reference to object containing data
            value: None or value the can be converted to bool
        """
        if value is not None:
            # Convert to string
            if not is_string(value):
                value = unicode(value)

        # String value
        instance.__dict__[self._name] = value


########################################


class StringListProperty(Property):
    """Class to enforce string list in member variable

Example:
"Inherit from (object) for Python 2.7"
>>> class foo(object):
    "Init to [\"foo\"]"
    x = StringListProperty("foo")
    "Init to None"
    y = StringListProperty()
    "Init to [\"a\",\"b\",\"c\"]"
    z = StringListProperty(["a","b","c"])

"Create the class"
f = foo()

"Print [\"foo\"]"
print(f.x)

"Print [\"False\"]"
f.x = "False"
print(f.x)

"Print True"
f.x = True
print(f.x)
    """

    def __get__(self, instance, owner=None):
        """Return value

        Args:
            instance: Reference to object containing data
            owner: Not used

        Returns:
            None, or verified data
        """

        # If never initialized, create an empty list
        result = instance.__dict__.get(self._name, None)
        if result is None:
            result = []
            # Make sure the value is set to the empty list
            instance.__dict__[self._name] = result
        return result

    def __set__(self, instance, value):
        """Set the string value
        Args:
            instance: Reference to object containing data
            value: None or value the can be converted to bool
        """

        if value is None:
            value = []
        else:
            # Convert to string
            if is_string(value) or not isinstance(value, Iterable):
                value = [unicode(value)]
            else:
                value = [unicode(i) for i in value]

        # String list value
        instance.__dict__[self._name] = value


########################################


class EnumProperty(Property):
    """
    Class to enforce string list in member variable

Attributes:
    _enums: Enumeration dictionary

Example:
j = (("a", "b", "c"), "d", "e", ["f", "g", "h"], "i")
"Inherit from (object) for Python 2.7"
>>> class foo(object):
    "Init to 0"
    x = EnumProperty(j, "a")
    "Init to 0"
    y = EnumProperty(j)
    "Init to 4"
    z = EnumProperty(j, "i")

"Create the class"
f = foo()

"Print 0"
print(f.x)

"Print 2"
f.x = "g"
print(f.x)

"Print 2"
f.x = "h"
print(f.x)
    """

    def __init__(self, name, enums):
        """Initialize to default
        Args:
            name: Name of the instance storage index
            enums: list of enumeration strings
        """

        self._enums = enums

        if not isinstance(enums, Iterable):
            raise ValueError(
                "enums \"{}\" is not an iterable".format(enums))

        if is_string(enums):
            raise ValueError(
                "enums \"{}\" can not be a string".format(enums))

        # Set the initial value using the derived class
        Property.__init__(self, name)

    def __set__(self, instance, value):
        """Set the string value
        Args:
            instance: Reference to object containing data
            value: None or value the can be converted to bool
        """

        if value is not None:
            if isinstance(value, Number):
                if value < 0:
                    raise ValueError(
                        "Value \"{}\" is less than zero".format(value))
                if value >= len(self._enums):
                    raise ValueError(
                        "Value {} is greater than or equal to {}".format(
                            value, len(self._enums)))
                value = int(value)
            else:
                # Check for an override
                enums = instance.__dict__.get(
                    self._name + "_enums", self._enums)

                # Convert to string
                for i, item in enumerate(enums):
                    if isinstance(item, Iterable):
                        if value in item:
                            value = i
                            break
                    if item == value:
                        value = i
                        break
                else:
                    raise ValueError(
                        "Value \"{}\" is not found in the list \"{}\"".format(
                            value, enums))

        # String list value
        instance.__dict__[self._name] = value

########################################


class NoneProperty(object):
    """
    Class to enforce None in member variable

Attributes:
    _name: The real name of the class instance

Example:
"Inherit from (object) for Python 2.7"
>>> class foo(object):
    "Init to None"
    x = NoneProperty()

"Create the class"
f = foo()

"Print None"
print(f.x)

"Exception on non None data"
f.x = "not None"
Traceback (most recent call last):
    ...
ValueError: Not None value
    """

    def __init__(self, name):
        """Initialize to default
        Args:
            name: Name of the instance storage index
        """

        self._name = name

    def __get__(self, instance, owner=None):
        """Return None

        Args:
            instance: Reference to object containing data
            owner: Not used

        Returns:
            None, or verified data
        """

        return None

    def __set__(self, instance, value):
        """Throw if not None

        Args:
            instance: Reference to object containing data
            value: None or value the can be converted to bool
        """

        if value is not None:
            raise ValueError(
                "\"{}\" can only be set to None, not \"{}\"".format(
                    self._name, value))
        instance.__dict__[self._name] = value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package that contains build helper functions

@package burger.buildutils

@var burger.buildutils._BURGER_SDKS_FOLDER
Cached location of the BURGER_SDKS folder

@var burger.buildutils._WINDOWS_ENV_PATHS
Environment variable locations of window applications
"""

# pylint: disable=consider-using-f-string
# pylint: disable=consider-using-with
# pylint: disable=import-error
# pylint: disable=deprecated-module

from __future__ import absolute_import, print_function, unicode_literals

import os
import platform
import subprocess
import sys
import errno

try:
    from wslwinreg import convert_from_windows_path
except ImportError:
    pass

from .strutils import is_string, encapsulate_path, get_windows_host_type, \
    get_mac_host_type, PY3_3_OR_HIGHER, PY3_5_OR_HIGHER, IS_CYGWIN, \
    IS_MSYS, IS_WSL, IS_WINDOWS, IS_WINDOWS_HOST

# Cached location of the BURGER_SDKS folder
_BURGER_SDKS_FOLDER = None

# Environment variable locations of window applications
_WINDOWS_ENV_PATHS = [
    "ProgramFiles",
    "ProgramFiles(x86)"
]

# For some goofy reason, Cygwin converts ProgramFiles
# into uppercase and preforms case sensitive comparisons
# To get around this, do the conversion for this table
if IS_CYGWIN or IS_MSYS:
    _WINDOWS_ENV_PATHS[0] = _WINDOWS_ENV_PATHS[0].upper()

########################################


def _create_header_guard(filename):
    """
    Create the header guard.

    Create the header guard by taking the filename,
    converting to upper case and replacing spaces and
    periods with underbars.

    Args:
        filename: Pathname of the file to generate a header guard.
    Returns:
        Base name, capitalized, and with underscores.
    """

    headerguard = os.path.basename(filename).upper()
    headerguard = headerguard.replace(" ", "_")
    return "__{}__".format(headerguard.replace(".", "_"))

########################################


def get_sdks_folder(verbose=False, refresh=False, folder=None):
    """
    Return the path of the BURGER_SDKS folder

    If the environment variable BURGER_SDKS is set,
    return the pathname it contains. Otherwise,
    print a warning if verbose is True and then attempt to find
    the "sdks" folder by traversing the current working directory
    for a folder named "sdks". If one isn't found, return None.

    Examples:
        # Normal use
        sdksfolder = burger.buildutils.get_sdks_folder()
        if not sdksfolder:
            print("failure")
            raise NameError("sdks not found, set BURGER_SDKS")

        # Alert the user if BURGER_SDKS isn't set
        burger.buildutils.get_sdks_folder(verbose=True)

        # Force the use of a supplied folder for sdks
        burger.buildutils.get_sdks_folder(refresh=True, folder="./foo/sdks/")

    Args:
        verbose: If True, print a message if BURGER_SDKS was not present
        refresh: If True, reset the cache and force a reload.
        folder: Path to use as BURGER_SDKS in the cache as an override

    Returns:
        None if the environment variable is not set, or the
        value of BURGER_SDKS.
    """

    global _BURGER_SDKS_FOLDER  # pylint: disable=global-statement

    # Clear the cache if needed
    if refresh:
        _BURGER_SDKS_FOLDER = None

    # Set the override, if found
    if folder:
        _BURGER_SDKS_FOLDER = folder

    # Not cached?
    if _BURGER_SDKS_FOLDER is None:

        # Load from the system
        _BURGER_SDKS_FOLDER = os.getenv("BURGER_SDKS", default=None)

        # Test for None or empty string
        if _BURGER_SDKS_FOLDER:
            _BURGER_SDKS_FOLDER = convert_from_windows_path(
                _BURGER_SDKS_FOLDER)

        else:
            # Warn about missing environment variable
            if verbose:
                print("The environment variable \"BURGER_SDKS\" is not set")

            # pylint: disable=import-outside-toplevel
            # Try to find the directory in the current path
            from .fileutils import traverse_directory
            sdks = traverse_directory(os.getcwd(), "sdks",
                                      find_directory=True, terminate=True)
            if sdks:
                _BURGER_SDKS_FOLDER = sdks[0]
                if verbose:
                    print(
                        "Assuming {} is the BURGER_SDKS folder".format(
                            sdks[0]))

    return _BURGER_SDKS_FOLDER

########################################


def fix_csharp(csharp_application_path):
    """
    Convert pathname to execute a C# exe file.

    @details C# applications can launch as is on Windows platforms,
    however, on Mac OSX and Linux, it must be launched
    from mono. Determine the host machine and if not
    windows, automatically prepend "mono" to
    the application's name to properly launch it

    This will also encase the name in quotes in case there are
    spaces in the pathname

    Args:
        csharp_application_path: Pathname string to update

    Returns:
        List of commands for the platform to launch a C# application.
    """

    # Prepend mono on non-windows systems
    if not get_windows_host_type(True):
        return ["mono", encapsulate_path(csharp_application_path)]
    return [csharp_application_path]

########################################


def is_exe(exe_path):
    """
    Return True if the file is executable

    Note:
        Windows platforms don't support the "x" bit so all
        files are executable if they exist.

    Args:
        exe_path: Full or partial pathname to test for existance
    Returns:
        True if the file is executable, False if the file doesn't exist or
        is not valid.
    """
    return os.path.isfile(exe_path) and os.access(exe_path, os.X_OK)

########################################


def get_path_ext(pathext=None):
    """
    Return a list of executable extensions

    If pathext is None, query the environment variable PATHEXT and
    return the entries as a string list. If pathext is a string,
    parse it as if it was a system specific PATHEXT string and
    if it's an iterable, return the value as is. If PATHEXT doesn't exist
    or is empty, return an empty list.

    Windows usually sets the value like this
    ``PATHEXT=.COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC``

    Args:
        pathext: String parsed as PATHEXT, iterable returned as is
    Returns:
        List of file name extension strings.
    See Also:
        burger.buildutils.make_exe_path, burger.buildutils.find_in_path
    """

    # Read the environment variable?
    if pathext is None:
        pathext = os.getenv("PATHEXT", None)
        if pathext is None:
            if IS_WSL:
                # Special case for WSL targets, allow .exe files
                return [".EXE"]
            return []

    # If a string, or environment variable?
    if is_string(pathext):
        # Special case for Cygwin, since os.pathsep
        # is ":" but the environment variable uses
        # ";" from Windows
        separator = ";" if IS_CYGWIN or IS_MSYS else os.pathsep
        # Parse the string
        pathext = pathext.split(separator)

    # Return the list or iterable
    return pathext

########################################


def make_exe_path(exe_path, pathext=None):
    """
    Given a folder and a executable name, return the true absolute path

    Examples:
        # exe could be returned as exe, exe.exe, exe.cmd, etc...
        path = make_exe_path("C:\\code\\exe")
        if path is None:
            print("No file named exe at C:\\code")

    Note:
        On macOS and Linux, PATHEXT is not set, this is for supporting
        extension types for common batch files or other executable extensions.

    Args:
        exe_path: Path of the executable to test
        pathext: Extension list to test
    Returns:
        None if a match was not found, or a full pathname with extension.

    See Also:
        burger.buildutils.get_path_ext, burger.buildutils.find_in_path
    """

    test_list = [exe_path]

    # Get the extension list
    pathext = get_path_ext(pathext)
    if pathext:
        # Only convert to lower case once
        exe_path_lower = exe_path.lower()

        # Does the file already have an extension?
        if not any(exe_path_lower.endswith(temp.lower()) for temp in pathext):
            # Create a list of possible file names with extensions
            test_list = [exe_path + temp for temp in pathext]

    # Try all the extensions (Can be an empty list)
    for temp_path in test_list:
        if is_exe(temp_path):
            break
    else:
        temp_path = None
    return temp_path

########################################


def find_in_path(filename, search_path=None, executable=False):
    """
    Using the system PATH environment variable, search for a file

    If the flag executable is False, the file will be found using a
    simple path search. If the flag is True, the file will be searched
    for using the extensions in the PATHEXT environment variable in
    addition to use the filename as is.

    If search_path is a string, it will be seperated using os.pathsep. If
    not, it will be treated as an interable list of strings of full pathnames
    to search. If it is None, the PATH environment variable will be used.

    Examples:
        # Can return "doxygen", "doxygen.exe" or "doxygen.com" depending
        # on what was found
        burger.find_in_path("doxygen", executable=True)

        # Will only find "foo.txt"
        burger.find_in_path("foo.txt")

    Args:
        filename: File to locate
        search_path: Search paths to use instead of PATH
        executable: True to ensure it's an executable
    Return:
        None if not found, a full path if the file is found.

    See Also:
        burger.buildutils.get_path_ext, burger.buildutils.make_exe_path
    """

    # pylint: disable=too-many-branches

    # Set up for added standard extentions
    test_list = [filename]
    if executable:

        # Are there path extensions (Windows)?
        pathext = get_path_ext()
        if pathext:

            # Only convert to lower case once
            filename_lower = filename.lower()

            # Does the file already have an extension?
            if not any(filename_lower.endswith(item.lower())
                       for item in pathext):
                # Create a list of possible file names with extensions
                test_list = [filename + item for item in pathext]

                # If Linux, allow "" as an extension
                if not IS_WINDOWS and IS_WINDOWS_HOST:
                    test_list.append(filename)

    # Is there a search path override?
    if not search_path:
        # Use the environment variable
        paths = os.getenv("PATH", "")
        if not paths:
            paths = os.defpath
    else:
        paths = search_path

    if is_string(paths):
        # Break it up based on the path separator
        paths = paths.split(os.pathsep)

    # On windows platforms, the current directory takes
    # precedence
    if search_path is None and get_windows_host_type(True):
        paths.insert(0, os.getcwd())

    # Scan the list of paths to find the file
    tested = set()
    for path in paths:
        # Normalize the path
        path = os.path.normcase(path)

        # Skip duplicates
        if path not in tested:
            tested.add(path)

            for item in test_list:
                temp_path = os.path.join(path, item)

                # Perform the test as an exe
                if executable:
                    if is_exe(temp_path):
                        break
                # Test for just a file
                elif os.path.isfile(temp_path):
                    break
            else:
                continue
            return os.path.normcase(os.path.normpath(temp_path))

    # Not found in the loops
    return None

########################################


def expand_and_verify(file_string):
    """
    Expand the input string with os.path.expandvars()

    After expanding the string, test for the existence of the file
    and return the expanded path if True. Otherwise, return None

    Examples:
        perforcepath = burger.expand_and_verify("${PERFORCE}\\p4.exe")
        if perforcepath is None:
            return

    Args:
        file_string: Pathname with environment variable tokens

    Returns:
        None if the string couldn't be expanded or if the file didn't exist,
            otherwise, return the expanded pathname

    """

    result_path = os.path.expandvars(file_string)
    if result_path is not None:
        result_path = convert_from_windows_path(result_path)
        if not os.path.isfile(result_path):
            result_path = None
    return result_path

########################################


def run_command(args, working_dir=None, quiet=False, capture_stdout=False,
                capture_stderr=False):
    """
    Execute a program and capture the return code and text output.

    Pass a command line formatted for the current shell and then this
    function will execute that command and capture both stdout and stderr if
    desired.

    Note:
        The first parameter is passed to subprocess.Popen() as is.

    Args:
        args: List of command line entries, starting with the program pathname
        working_dir: Directory to set before executing command
        quiet: Set to True if errors should not be printed
        capture_stdout: Set to True if stdout is to be captured
        capture_stderr: Set to True if stderr is to be captured
    Returns:
        The return error_code, stdout, stderr
    """

    # Which output streams are to be captured?
    stdout = subprocess.PIPE if capture_stdout else None
    stderr = subprocess.PIPE if capture_stderr else None

    # Set the working directory if needed
    old_directory = os.getcwd()
    if working_dir:
        os.chdir(working_dir)

    try:
        tempfp = subprocess.Popen(args, cwd=working_dir, stdout=stdout,
                                  stderr=stderr, universal_newlines=True)
    except OSError as error:
        if not quiet:
            if is_string(args):
                msg = args
            else:
                msg = " ".join(args)
            print("Command line \"{}\" generated error {}".format(msg, error))

        # Restore directory
        os.chdir(old_directory)
        return (error.errno, "", "")

    # Restore directory
    os.chdir(old_directory)

    stdoutstr, stderrstr = tempfp.communicate()
    return (tempfp.returncode, stdoutstr, stderrstr)

########################################


def is_codewarrior_mac_allowed():
    """
    Return True if this machine can run Codewarrior for Mac OS Carbon.

    Test first if the host platform is a mac, and if so, test if it's
    capable of running Mac OS Carbon Codewarrior 9 or 10

    Returns:
        True if CodeWarrior for Mac OS can be run on this Macintosh

    See Also:
        strutils.host_machine
    """

    # Test if a mac

    if get_mac_host_type():
        # Get the Mac OS version number
        mac_ver = platform.mac_ver()
        release = mac_ver[0]

        # Convert 10.5.8 to 10.5

        digits = release.split(".")

        # Snow Leopard (10.6) supports Rosetta
        # Lion (10.7) and Mountain Lion (10.8) do not

        if float(digits[0]) >= 10:
            if float(digits[1]) < 7:
                return True

    # Can't run, not a mac or Power PC native or emulation isn't supported
    return False

########################################


def import_py_script(file_name, module_name=None):
    """
    Manually load in a python file.

    Load in a python script from disk and parse it, creating
    a .pyc file if needed and reading from a .pyc file if it exists.

    Note:
        The module returned will not be present in the sys.modules cache, this
        is by design to allow python files with the same name to be loaded
        from different directories without creating a cache collision

    Args:
        file_name: Name of the file to load
        module_name: Name of the loaded module for ``__name__``
    Returns:
        The imported python script object
    See Also:
        run_py_script
    """

    # pylint: disable=import-outside-toplevel
    # pylint: disable=too-many-branches

    # If there's no module name, glean one from the filename
    if not module_name:
        module_name = os.path.splitext(os.path.split(file_name)[-1])[0]

    old_dont = sys.dont_write_bytecode
    sys.dont_write_bytecode = True
    try:
        if PY3_5_OR_HIGHER:

            # Python 3.5 and allows the loading of a module without
            # touching the cache
            import importlib.util
            spec = importlib.util.spec_from_file_location(
                module_name, file_name)

            # File not found?
            if not spec:
                result = None
            else:
                # Import and then execute
                result = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(result)

        else:
            # First step, if there's a module already loaded by this
            # name, save it for restoration later

            saved = None
            if module_name in sys.modules:
                saved = sys.modules[module_name]
                del sys.modules[module_name]

            # Perform the load, throw exception on error
            try:
                if PY3_3_OR_HIGHER:
                    # Python 3.3 and 3.4 prefers using the SourceFileLoader
                    # class
                    from importlib.machinery import SourceFileLoader
                    # pylint: disable=deprecated-method
                    result = SourceFileLoader(
                        module_name, file_name).load_module(None)

                else:
                    # Use the imp library for Python 2.x to 3.2
                    import imp  # type: ignore
                    result = imp.load_source(module_name, file_name)

            # Wrap up by restoring the cache the way it was found
            finally:
                if saved:
                    sys.modules[module_name] = saved
                else:
                    # Remove the generated entry since load_source() added it

                    # Note: Test before deletion, in case load_source threw
                    # an exception before creating the entry
                    if module_name in sys.modules:
                        del sys.modules[module_name]

    except IOError as error:
        # File not found is the correct error
        if error.errno == errno.ENOENT:
            result = None
        else:
            raise

    finally:
        sys.dont_write_bytecode = old_dont

    return result

########################################


def run_py_script(file_name, function_name=None, arg=None):
    """
    Manually load and run a function in a python file.

    Load in a python script from disk and execute a specific function.
    Returns the value returned from the loaded script.

    Note:
        The script will not be added to the module cache.

    Args:
        file_name: Name of the file to load
        function_name: Name of the function in the file to call
        arg: Argument to pass to the function
    Returns:
        The value returned from the python script.
    See Also:
        import_py_script
    """

    # If a function name wasn't passed, assume it's ``main``
    if not function_name:
        function_name = "main"

    # Load in the script
    module = import_py_script(file_name)

    # Find the function and execute it
    method = getattr(module, function_name)
    if arg is None:
        return method()
    return method(arg)

# pylint: disable=redefined-builtin


def execfile(filename, globals, locals=None):
    """
    Implementation of execfile from Python 2

    It's not exact. This version requires a globals object.

    To maintain compatilibity to execfile() in python 2, the
    input parameters are hard coded to filename, globals,
    and locals despite what pylint insists on.

    Args:
        filename: Full path of python file to load and execute
        globals: Usually globals()
        locals: Optional, usually locals()

    Returns:
        None
    """

    # pylint: disable=exec-used

    # If no locals are passed, use the globals
    if locals is None:
        locals = globals

    # Read in the python file
    with open(filename, "rb") as inputfile:
        source = inputfile.read()

    # Compile to byte code
    bytecode = compile(source, filename, "exec")
    exec(bytecode, globals, locals)
